__all__ = ["JSONArtifact"]

from pathlib import Path
from typing import Any

from gravitorch.utils.artifacts.base import BaseArtifact
from gravitorch.utils.io import save_json


class JSONArtifact(BaseArtifact):
    r"""Implements a JSON artifact.

    Args:
        tag (str): Specifies the artifact tag. The tag is used to
            define the JSON filename.
        data: Specifies the data to save in the JSON artifact. The data
            should be JSON compatible.
    """

    def __init__(self, tag: str, data: Any):
        self._tag = str(tag)
        self._data = data

    def __str__(self) -> str:
        return f"{self.__class__.__qualname__}(tag={self._tag})"

    def create(self, path: Path) -> None:
        save_json(to_save=self._data, path=path.joinpath(f"{self._tag}.json"))



__all__ = ["BaseResource", "setup_resource"]

import logging
from contextlib import AbstractContextManager
from typing import Union

from objectory import AbstractFactory

from gravitorch.utils.format import str_target_object

logger = logging.getLogger(__name__)


class BaseResource(AbstractContextManager, metaclass=AbstractFactory):
    r"""Defines the base class to manage a resource."""


def setup_resource(resource: Union[BaseResource, dict]) -> BaseResource:
    r"""Sets up a resource.

    The resource manager is instantiated from its configuration by using the
    ``BaseResource`` factory function.

    Args:
        resource (``BaseResource`` or dict): Specifies
            the resource or its configuration.

    Returns:
        ``BaseResource``: The instantiated resource.
    """
    if isinstance(resource, dict):
        logger.debug(
            f"Initializing a resource from its configuration... {str_target_object(resource)}"
        )
        resource = BaseResource.factory(**resource)
    return resource



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['gravitorch',
 'gravitorch.cli',
 'gravitorch.creators',
 'gravitorch.creators.core',
 'gravitorch.creators.dataloader',
 'gravitorch.creators.datapipe',
 'gravitorch.creators.datasource',
 'gravitorch.creators.lr_scheduler',
 'gravitorch.creators.model',
 'gravitorch.creators.optimizer',
 'gravitorch.data',
 'gravitorch.data.datacreators',
 'gravitorch.data.dataloaders',
 'gravitorch.data.dataloaders.collators',
 'gravitorch.data.datapipes',
 'gravitorch.data.datapipes.iter',
 'gravitorch.data.datasets',
 'gravitorch.data.partitioners',
 'gravitorch.datasources',
 'gravitorch.distributed',
 'gravitorch.engines',
 'gravitorch.experimental',
 'gravitorch.handlers',
 'gravitorch.lr_schedulers',
 'gravitorch.models',
 'gravitorch.models.criterions',
 'gravitorch.models.metrics',
 'gravitorch.models.metrics.classification',
 'gravitorch.models.metrics.regression',
 'gravitorch.models.networks',
 'gravitorch.models.utils',
 'gravitorch.nn',
 'gravitorch.nn.experimental',
 'gravitorch.nn.functional',
 'gravitorch.nn.functional.experimental',
 'gravitorch.nn.fusion',
 'gravitorch.nn.utils',
 'gravitorch.optimizers',
 'gravitorch.rsrc',
 'gravitorch.runners',
 'gravitorch.utils',
 'gravitorch.utils.artifacts',
 'gravitorch.utils.data_summary',
 'gravitorch.utils.device_placement',
 'gravitorch.utils.engine_states',
 'gravitorch.utils.evaluation_loops',
 'gravitorch.utils.events',
 'gravitorch.utils.exp_trackers',
 'gravitorch.utils.history',
 'gravitorch.utils.loop_observers',
 'gravitorch.utils.meters',
 'gravitorch.utils.parameter_initializers',
 'gravitorch.utils.profilers',
 'gravitorch.utils.tensor',
 'gravitorch.utils.training_loops']

package_data = \
{'': ['*']}

install_requires = \
['coola>=0.0,<0.1',
 'hya>=0.0,<0.1',
 'hydra-core>=1.3,<2.0',
 'numpy>=1.23,<2.0',
 'objectory>=0.0,<0.1',
 'psutil>=5.9,<6.0',
 'pytorch-ignite>=0.4,<0.5',
 'tabulate>=0.9,<0.10',
 'torch>=1.13,<2.0',
 'tqdm>=4.64,<5.0']

extras_require = \
{'all': ['accelerate>=0.15,<0.16',
         'colorlog>=6.7,<7.0',
         'matplotlib>=3.6,<4.0',
         'pillow>=9.3,<10.0',
         'tensorboard>=2.11,<3.0',
         'torchvision>=0.14,<0.15'],
 'tb': ['tensorboard>=2.11,<3.0'],
 'vision': ['pillow>=9.3,<10.0', 'torchvision>=0.14,<0.15']}

setup_kwargs = {
    'name': 'gravitorch',
    'version': '0.0.4',
    'description': 'Warning: API is not stable',
    'long_description': '\n[//]: # (<p align="center">)\n\n[//]: # (   <a href="https://github.com/durandtibo/meteor/actions">)\n\n[//]: # (      <img alt="CI" src="https://github.com/durandtibo/meteor/workflows/CI/badge.svg?event=push&branch=main">)\n\n[//]: # (   </a>)\n\n[//]: # (    <a href="https://pypi.org/project/gravitorch/">)\n\n[//]: # (      <img alt="PYPI version" src="https://img.shields.io/pypi/v/gravitorch">)\n\n[//]: # (    </a>)\n\n[//]: # (   <a href="https://pypi.org/project/gravitorch/">)\n\n[//]: # (      <img alt="Python" src="https://img.shields.io/pypi/pyversions/gravitorch.svg">)\n\n[//]: # (   </a>)\n\n[//]: # (   <a href="https://opensource.org/licenses/BSD-3-Clause">)\n\n[//]: # (      <img alt="BSD-3-Clause" src="https://img.shields.io/pypi/l/gravitorch">)\n\n[//]: # (   </a>)\n\n[//]: # (   <a href="https://codecov.io/gh/durandtibo/meteor">)\n\n[//]: # (      <img alt="Codecov" src="https://codecov.io/gh/durandtibo/meteor/branch/main/graph/badge.svg">)\n\n[//]: # (   </a>)\n\n[//]: # (   <a href="https://github.com/psf/black">)\n\n[//]: # (     <img  alt="Code style: black" src="https://img.shields.io/badge/code%20style-black-000000.svg">)\n\n[//]: # (   </a>)\n\n[//]: # (   <a href="https://google.github.io/styleguide/pyguide.html#s3.8-comments-and-docstrings">)\n\n[//]: # (     <img  alt="Doc style: google" src="https://img.shields.io/badge/%20style-google-3666d6.svg">)\n\n[//]: # (   </a>)\n\n[//]: # (   <br/>)\n\n[//]: # (</p>)\n\n\n<p align="center">\n   <a>\n      <img alt="CI" src="https://github.com/durandtibo/meteor/workflows/CI/badge.svg?event=push&branch=main">\n   </a>\n    <a href="https://pypi.org/project/gravitorch/">\n      <img alt="PYPI version" src="https://img.shields.io/pypi/v/gravitorch">\n    </a>\n   <a href="https://pypi.org/project/gravitorch/">\n      <img alt="Python" src="https://img.shields.io/pypi/pyversions/gravitorch.svg">\n   </a>\n   <a href="https://opensource.org/licenses/BSD-3-Clause">\n      <img alt="BSD-3-Clause" src="https://img.shields.io/pypi/l/gravitorch">\n   </a>\n   <a>\n      <img alt="Codecov" src="https://codecov.io/gh/durandtibo/meteor/branch/main/graph/badge.svg">\n   </a>\n   <a href="https://github.com/psf/black">\n     <img  alt="Code style: black" src="https://img.shields.io/badge/code%20style-black-000000.svg">\n   </a>\n   <a href="https://google.github.io/styleguide/pyguide.html#s3.8-comments-and-docstrings">\n     <img  alt="Doc style: google" src="https://img.shields.io/badge/%20style-google-3666d6.svg">\n   </a>\n   <br/>\n</p>\n',
    'author': 'Thibaut Durand',
    'author_email': 'durand.tibo+gh@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



r"""This package contains the implementation of some engine states."""

__all__ = ["BaseEngineState", "VanillaEngineState", "setup_engine_state"]

from gravitorch.utils.engine_states.base import BaseEngineState
from gravitorch.utils.engine_states.utils import setup_engine_state
from gravitorch.utils.engine_states.vanilla import VanillaEngineState



__all__ = ["VanillaDataSourceCreator"]

import logging

from gravitorch import constants as ct
from gravitorch.creators.datasource.base import BaseDataSourceCreator
from gravitorch.datasources.base import BaseDataSource
from gravitorch.datasources.utils import setup_and_attach_data_source, setup_data_source
from gravitorch.engines.base import BaseEngine

logger = logging.getLogger(__name__)


class VanillaDataSourceCreator(BaseDataSourceCreator):
    r"""Implements a simple data source creator.

    Args:
        config (dict): Specifies the data source configuration.
        attach_to_engine (bool, optional): If ``True``, the data
            source is attached to the engine. Default: ``True``
        add_module_to_engine (bool, optional): If ``True``, the data
            source is added to the engine state, so the data source
            state is stored when the engine creates a checkpoint.
            Default: ``True``
    """

    def __init__(
        self,
        config: dict,
        attach_to_engine: bool = True,
        add_module_to_engine: bool = True,
    ):
        self._config = config
        self._attach_to_engine = bool(attach_to_engine)
        self._add_module_to_engine = bool(add_module_to_engine)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(attach_to_engine={self._attach_to_engine}, "
            f"add_module_to_engine={self._add_module_to_engine})"
        )

    def create(self, engine: BaseEngine) -> BaseDataSource:
        logger.info("Creating a data source...")
        if self._attach_to_engine:
            data_source = setup_and_attach_data_source(data_source=self._config, engine=engine)
        else:
            data_source = setup_data_source(data_source=self._config)
        if self._add_module_to_engine:
            logger.info(f"Adding a data source to the engine (key: {ct.DATA_SOURCE})...")
            engine.add_module(ct.DATA_SOURCE, data_source)
        return data_source



r"""This module implements some utility functions to use ``torch.Tensor``s."""

__all__ = [
    "has_name",
    "partial_transpose_dict",
    "permute",
    "shapes_are_equal",
    "str_full_tensor",
    "to_tensor",
]

from collections.abc import Sequence
from typing import Any, TypeVar

import numpy as np
import torch
from torch import Tensor

T = TypeVar("T")


def str_full_tensor(tensor: Tensor) -> str:
    r"""Computes a string representation of the tensor with all the values.

    Args:
        tensor (``torch.Tensor``): Specifies the input tensor.

    Returns:
         str: The string representation of the tensor with all the
            values.
    """
    torch.set_printoptions(profile="full")
    output = str(tensor)
    torch.set_printoptions(profile="default")  # reset
    return output


def has_name(tensor: Tensor) -> bool:
    r"""Indicates it the tensor has at least one name or not.

    Args:
        tensor (``torch.Tensor``): Specifies the tensor to check.

    Returns:
        bool: ``True`` if tensor has at least one name, otherwise
            ``False``.

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.tensor import has_name
        >>> has_name(torch.ones(2, 3))
        False
        >>> has_name(torch.ones(2, 3, names=('B', None)))
        True
        >>> has_name(torch.ones(2, 3, names=('B', 'F')))
        True
    """
    return set(tensor.names) != {None}


def permute(tensor: Tensor, permutation: Tensor, dim: int = 0) -> Tensor:
    r"""Permutes a tensor on a given dimension and a permutation.

    Args:
        tensor (``torch.Tensor``): Specifies the tensor to permute.
        permutation (``torch.Tensor`` of type long and shape
            ``(dimension,)``): Specifies the permutation to use on the
            tensor. The dimension of this tensor should be compatible
            with the shape of the tensor to permute.
        dim (int, optional): Specifies the dimension used to permute the
            tensor. Default: ``0``

    Returns:
        ``torch.Tensor``: The permuted tensor.

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.tensor import permute
        >>> permute(tensor=torch.arange(4), permutation=torch.tensor([0, 2, 1, 3]))
        tensor([0, 2, 1, 3])
        >>> permute(tensor=torch.arange(20).view(4, 5), permutation=torch.tensor([0, 2, 1, 3]))
        tensor([[0, 1, 2, 3, 4], [10, 11, 12, 13, 14], [5, 6, 7, 8, 9], [15, 16, 17, 18, 19]])
        >>> permute(
        ...     tensor=torch.arange(20).view(4, 5),
        ...     permutation=torch.tensor([0, 4, 2, 1, 3]),
        ...     dim=1,
        ... )
        tensor([[ 0,  4,  2,  1,  3],
                [ 5,  9,  7,  6,  8],
                [10, 14, 12, 11, 13],
                [15, 19, 17, 16, 18]])
        >>> permute(
        ...     tensor=torch.arange(20).view(2, 2, 5),
        ...     permutation=torch.tensor([0, 4, 2, 1, 3]),
        ...     dim=2,
        ... )
        tensor([[[ 0,  4,  2,  1,  3],
                 [ 5,  9,  7,  6,  8]],
                [[10, 14, 12, 11, 13],
                 [15, 19, 17, 16, 18]]])
    """
    return tensor.transpose(0, dim)[permutation].transpose(0, dim).contiguous()


def partial_transpose_dict(data: dict, config: dict) -> dict:
    r"""Transposes some ``torch.Tensor``s that are in a dictionary.

    Note the transposed tensors may not be contiguous in memory
    tensors.

    Args:
        data (dict): Specifies the dictionary with the tensors to
            transpose.
        config (dict): Specifies the tensors to transpose. The keys
            should exist in the ``data`` input. The keys indicate the
            tensors to transpose, and the values indicate the
            dimension to transpose. See example below for more
            details.

    Returns:
        dict: A dictionary with some transposed tensors.

    Example usage:

    .. code-block:: python

        >>> x = torch.arange(10).view(2, 5)
        >>> x
        tensor([[0, 1, 2, 3, 4],
                [5, 6, 7, 8, 9]])
        # No transposition
        >>> from gravitorch.utils.tensor import partial_transpose_dict
        >>> partial_transpose_dict({'my_key': x}, {})
        {'my_key': tensor([[0, 1, 2, 3, 4],
                 [5, 6, 7, 8, 9]])}
        # Transpose the first two dimensions
        >>> partial_transpose_dict({'my_key': x}, {'my_key': [0, 1]})
        {'my_key': tensor([[0, 5],
                 [1, 6],
                 [2, 7],
                 [3, 8],
                 [4, 9]])}
        # The order of the dimensions is not important  [0, 1] <> [1, 0]
        >>> partial_transpose_dict({'my_key': x}, {'my_key': [1, 0]})
        {'my_key': tensor([[0, 5],
                 [1, 6],
                 [2, 7],
                 [3, 8],
                 [4, 9]])}
        >>> x = torch.arange(24).view(2, 3, 4)
        >>> x
        tensor([[[ 0,  1,  2,  3],
                 [ 4,  5,  6,  7],
                 [ 8,  9, 10, 11]],

                [[12, 13, 14, 15],
                 [16, 17, 18, 19],
                 [20, 21, 22, 23]]])
        >>> partial_transpose_dict({'my_key': x}, {'my_key': [0, 1]})
        {'my_key': tensor([[[ 0,  1,  2,  3],
                  [12, 13, 14, 15]],

                 [[ 4,  5,  6,  7],
                  [16, 17, 18, 19]],

                 [[ 8,  9, 10, 11],
                  [20, 21, 22, 23]]])}
        >>> partial_transpose_dict({'my_key': x}, {'my_key': [0, 2]})
        {'my_key': tensor([[[ 0, 12],
                  [ 4, 16],
                  [ 8, 20]],

                 [[ 1, 13],
                  [ 5, 17],
                  [ 9, 21]],

                 [[ 2, 14],
                  [ 6, 18],
                  [10, 22]],

                 [[ 3, 15],
                  [ 7, 19],
                  [11, 23]]])}
        >>> partial_transpose_dict({'my_key': x}, {'my_key': [1, 2]})
        {'my_key': tensor([[[ 0,  4,  8],
                  [ 1,  5,  9],
                  [ 2,  6, 10],
                  [ 3,  7, 11]],

                 [[12, 16, 20],
                  [13, 17, 21],
                  [14, 18, 22],
                  [15, 19, 23]]])}
    """
    for key, dims in config.items():
        if key not in data:
            raise ValueError(f"key '{key}' is not in the data dict (keys: {list(data.keys())})")
        data[key] = data[key].transpose(*dims)
    return data


def to_tensor(value: Any) -> Tensor:
    r"""Converts the input to a ``torch.Tensor``.

    Args:
        value: Specifies the input to convert to ``torch.Tensor``.

    Returns:
        ``torch.Tensor``: The tensor.

    Raises:
        TypeError if the type cannot be converted to ``torch.Tensor``.

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.utils.tensor import to_tensor
        >>> to_tensor(torch.tensor([-3, 1, 7]))
        tensor([-3,  1,  7])
        >>> to_tensor([-3, 1, 7])
        tensor([-3,  1,  7])
        >>> to_tensor((-3, 1, 7))
        tensor([-3,  1,  7])
        >>> to_tensor(np.array([-3, 1, 7]))
        tensor([-3,  1,  7])
    """
    if torch.is_tensor(value):
        return value
    if isinstance(value, np.ndarray):
        return torch.from_numpy(value)
    if isinstance(value, Sequence):
        return torch.tensor(value)
    if isinstance(value, (int, float)):
        return torch.tensor(value)
    raise TypeError(f"Incorrect type: {type(value)}")


def shapes_are_equal(tensors: Sequence[Tensor]) -> bool:
    r"""Indicates if the shapes of several tensors are equal or not.

    This method does not check the values or the data type of the
    tensors.

    Args:
        tensors (sequence): Specifies the tensors to check.

    Returns:
        bool: ``True`` if all the tensors have the same shape,
            otherwise ``False``. By design, this function returns
            ``False`` if no tensor is provided.

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.utils.tensor import shapes_are_equal
        >>> shapes_are_equal([torch.rand(2, 3), torch.rand(2, 3)])
        True
        >>> shapes_are_equal([torch.rand(2, 3), torch.rand(2, 3, 1)])
        False
    """
    if not tensors:
        return False
    shape = tensors[0].shape
    for tensor in tensors[1:]:
        if shape != tensor.shape:
            return False
    return True



__all__ = ["BaseLRSchedulerCreator", "VanillaLRSchedulerCreator", "setup_lr_scheduler_creator"]

from gravitorch.creators.lr_scheduler.base import BaseLRSchedulerCreator
from gravitorch.creators.lr_scheduler.utils import setup_lr_scheduler_creator
from gravitorch.creators.lr_scheduler.vanilla import VanillaLRSchedulerCreator



r"""This module implements PyTorch sampler functions for data loaders."""

__all__ = ["ReproducibleBatchSampler", "PartialSequentialSampler", "PartialRandomSampler"]

from collections.abc import Generator, Iterator

import torch
from torch.utils.data import BatchSampler, Sampler


class ReproducibleBatchSampler(BatchSampler):
    r"""Implements a reproducible batch sampler.

    This class is inspired from PyTorch Ignite. Internally, this class
    iterates and stores indices of the input batch sampler.

    Args:
        batch_sampler (``torch.utils.data.BatchSampler``): batch
            sampler same as used with ``torch.utils.data.DataLoader``
        start_iteration (int, optional): Specifies the starting
            iteration. Default: 0.

    Example usage:

    .. code-block:: python

        >>> from gravitorch.data.dataloaders.samplers import ReproducibleBatchSampler
        >>> from torch.utils.data import BatchSampler, SequentialSampler
        >>> batch_sampler = BatchSampler(
        ...     SequentialSampler(range(10)),
        ...     batch_size=3,
        ...     drop_last=False,
        ... )
        >>> reproducible_batch_sampler = ReproducibleBatchSampler(batch_sampler)
        >>> list(reproducible_batch_sampler)
        [[0, 1, 2], [3, 4, 5], [6, 7, 8], [9]]
        >>> batch_sampler = ReproducibleBatchSampler(batch_sampler, start_iteration=1)
        >>> list(reproducible_batch_sampler)
        [[3, 4, 5], [6, 7, 8], [9]]
        >>> batch_sampler = ReproducibleBatchSampler(batch_sampler, start_iteration=2)
        >>> list(reproducible_batch_sampler)
        [[6, 7, 8], [9]]
    """

    def __init__(self, batch_sampler: BatchSampler, start_iteration: int = 0):
        if not isinstance(batch_sampler, BatchSampler):
            raise TypeError(
                "Argument batch_sampler should be torch.utils.data.sampler.BatchSampler"
            )
        if start_iteration < 0:
            raise ValueError("Argument start_iteration should be positive integer")

        self.batch_sampler = batch_sampler
        self._start_iteration = start_iteration

    def __iter__(self) -> Generator:
        yield from list(self.batch_sampler)[self._start_iteration :]

    def __len__(self) -> int:
        return len(self.batch_sampler)


class PartialSequentialSampler(Sampler):
    r"""Implements a partial sequential sampler that samples only the first
    items in the dataset.

    Args:
        data_source (``torch.utils.data.Dataset``): Specifies the
            dataset to sample from.
        num_samples (int): Specifies the number of samples to draw.
            If the number of samples is bigger than the number of
            samples in the dataset, the number of samples to draw
            is the dataset size.
    """

    def __init__(self, data_source, num_samples: int):
        super().__init__(data_source)
        self.data_source = data_source
        if not isinstance(num_samples, int) or num_samples <= 0:
            raise ValueError(
                f"num_samples should be a positive integer value, but got {num_samples}"
            )
        self.num_samples = num_samples

    def __iter__(self) -> Iterator:
        return iter(range(len(self)))

    def __len__(self) -> int:
        return min(len(self.data_source), self.num_samples)


class PartialRandomSampler(PartialSequentialSampler):
    r"""Implements a partial random sampler that samples randomly some items of
    the dataset.

    Args:
        data_source (``torch.utils.data.Dataset``): Specifies the
            dataset to sample from.
        nnum_samples (int): Specifies the number of samples to draw.
            If the number of samples is bigger than the number of
            samples in the dataset, the number of samples to draw
            is the dataset size.
    """

    def __iter__(self) -> Iterator:
        return iter(torch.randperm(len(self.data_source))[: self.num_samples].tolist())



__all__ = ["EpochRandomIterDataPipeCreator"]

import logging
from collections.abc import Sequence
from typing import Optional

from objectory import OBJECT_TARGET, factory
from torch.utils.data import IterDataPipe

from gravitorch import distributed as dist
from gravitorch.creators.datapipe.base import BaseIterDataPipeCreator
from gravitorch.engines.base import BaseEngine

logger = logging.getLogger(__name__)


class EpochRandomIterDataPipeCreator(BaseIterDataPipeCreator):
    r"""Implements an ``IterDataPipe`` creator to create an ``IterDataPipe``
    object where its random seed is controlled by an engine.

    Given an engine, the random seed is set based on the engine random
    seed, the current epoch value, the maximum number of epochs and
    the distributed rank.

    Args:
        config (dict): Specifies the configuration of the
            ``IterDataPipe``.
        random_seed_key (str, optional): Specifies the key in the
            configuration which is used to indicate the random seed
            of the ``IterDataPipe``. If this key exists in ``config``,
            it will be replaced by a new value, based on the engine
            state.
    """

    def __init__(self, config: dict, random_seed_key: str = "random_seed"):
        self._config = config
        self._random_seed_key = str(random_seed_key)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(config={self._config}, "
            f"random_seed_key={self._random_seed_key})"
        )

    def create(
        self, engine: Optional[BaseEngine] = None, source_inputs: Optional[Sequence] = None
    ) -> IterDataPipe:
        if engine is None:
            raise ValueError(
                "engine cannot be None because the epoch value is used to create the "
                "IterDataPipe object"
            )
        source_inputs = source_inputs or tuple()
        config = self._config.copy()  # Make a copy because the dict is modified below.
        target = config.pop(OBJECT_TARGET)
        config[self._random_seed_key] = (
            config.get(self._random_seed_key, engine.random_seed)
            + engine.epoch
            + engine.max_epochs * dist.get_rank()
        )
        logger.info(
            f"Set the random seed to {config[self._random_seed_key]}  [{OBJECT_TARGET}: {target}]"
        )
        return factory(target, *source_inputs, **config)



r"""This module implements a wrapper network to make image classification
network from ``torchvision`` compatible with ``VanillaModel``."""

__all__ = ["ImageClassificationNetwork"]

from typing import Any, Union

import torch
from torch import Tensor
from torch.nn import Module

from gravitorch import constants as ct
from gravitorch.nn.utils.factory import setup_nn_module
from gravitorch.nn.utils.module_helpers import get_module_device


class ImageClassificationNetwork(Module):
    r"""Implements a wrapper network for image classification network from
    ``torchvision``.

    This network assumes that the input shape is ``224*224``.

    Args:
        network (``torch.nn.Module`` or dict): Specifies the network
            module or its configuration.
        input_name (str, optional): Specifies the input name.
            Default: ``'input'``
        output_name (str, optional): Specifies the output name.
            Default: ``'prediction'``
    """

    def __init__(
        self,
        network: Union[Module, dict[str, Any]],
        input_name: str = ct.INPUT,
        output_name: str = ct.PREDICTION,
    ):
        super().__init__()
        self.network = setup_nn_module(network)
        self._input_name = input_name
        self._output_name = output_name

    def forward(self, tensor: Tensor) -> Tensor:
        r"""Computes the predictions of the network.

        Args:
            tensor (``torch.Tensor`` of shape
                ``(batch size, num channels, height, width)`` and
                type float): Specifies the input images.

        Returns:
            ``torch.Tensor`` of shape ``(batch size, num classes)``
                and type float: The prediction for each class.
        """
        return self.network(tensor)

    def get_dummy_input(self, batch_size: int = 1) -> tuple[Tensor]:
        r"""Generates a dummy input for the MLP.

        Args:
            batch_size (int, optional): Specifies the batch size to
                use to generate the dummy input. Default: ``1``

        Returns:
            ``tuple[Tensor]``: A tuple with one tensor of shape
                ``(batch_size, 3, 244, 244)`` and type ``float``.
                The tensor is on the same device that this network.
        """
        return (torch.randn(batch_size, 3, 224, 224, device=get_module_device(self.network)),)

    def get_input_names(self) -> tuple[str]:
        r"""Gets the tuple of input names for the network.

        The order of the name should be the same that the order in
        the inputs of the forward function.

        Returns:
            tuple: The tuple of input names.
        """
        return (self._input_name,)

    def get_onnx_dynamic_axis(self) -> dict:
        r"""Gets the dynamic axes (e.g. batch size or sequence length) when
        serializing a model to ONNX format.

        See https://pytorch.org/docs/stable/onnx.html#torch.onnx.export
        to have more information on how to create the ``dict``.

        Returns:
            dict: with the dynamic axes of the input/output.
        """
        return {self._input_name: {0: "batch"}, self._output_name: {0: "batch"}}

    def get_output_names(self) -> tuple[str]:
        r"""Gets the tuple of output names for the network.

        The order of the name should be the same that the order in
        the outputs of the forward function.

        Returns:
            tuple: The tuple of output names.
        """
        return (self._output_name,)



r"""This module implements a data summary for float values."""

__all__ = ["FloatDataSummary"]

from typing import TypeVar

from gravitorch.utils.data_summary.continuous import BaseContinuousDataSummary

T = TypeVar("T")

DEFAULT_QUANTILES = (0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0)


class FloatDataSummary(BaseContinuousDataSummary[float]):
    r"""Implements a data summary for float values.

    This data summary assumes that the data are continuous numerical
    values. This data summary computes the following statistics:

        - ``count``: the number of values
        - ``sum``: the sum of tne values
        - ``mean``: the mean of tne values
        - ``median``: the median of tne values
        - ``std``: the standard deviation of the values
        - ``max``: the max value
        - ``min``: the min value
        - ``quantiles``: the quantile values

    Args:
        max_size (int, optional): Specifies the maximum size used to
            store the last values because it may not be possible to
            store all the values. This parameter is used to compute
            the median and the quantiles. Default: ``10000``
        quantiles (``torch.Tensor`` or tuple or list, optional):
            Specifies a sequence of quantiles to compute, which must
            be between 0 and 1 inclusive. Default:
            ``(0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0)``
    """

    def add(self, data: float) -> None:
        r"""Adds new data to the summary.

        Args:
            data (float): Specifies the data to add to the summary.
        """
        value = float(data)
        self._sum += value
        self._count += 1.0
        self._min_value = min(self._min_value, value)
        self._max_value = max(self._max_value, value)
        self._values.append(value)



__all__ = [
    "BaseLRScheduler",
    "InverseSquareRootLR",
    "LRSchedulerType",
    "create_linear_warmup_cosine_decay_lr",
    "create_linear_warmup_linear_decay_lr",
    "create_sequential_lr",
    "setup_lr_scheduler",
]

from gravitorch.lr_schedulers.base import (
    BaseLRScheduler,
    LRSchedulerType,
    setup_lr_scheduler,
)
from gravitorch.lr_schedulers.creation import (
    create_linear_warmup_cosine_decay_lr,
    create_linear_warmup_linear_decay_lr,
    create_sequential_lr,
)
from gravitorch.lr_schedulers.invsqrt import InverseSquareRootLR



__all__ = [
    "create_linear_warmup_cosine_decay_lr",
    "create_linear_warmup_linear_decay_lr",
    "create_sequential_lr",
]

from collections.abc import Sequence
from typing import Union

from torch.optim import Optimizer
from torch.optim.lr_scheduler import CosineAnnealingLR, LinearLR, SequentialLR

from gravitorch.lr_schedulers.base import LRSchedulerType, setup_lr_scheduler


def create_sequential_lr(
    optimizer: Optimizer,
    schedulers: Sequence[Union[LRSchedulerType, dict]],
    milestones: list[int],
    last_epoch: int = -1,
) -> SequentialLR:
    r"""Instantiates a sequential LR scheduler from its configuration.

    Args:
        optimizer (``torch.optim.Optimizer``): Specifies the optimizer
            associated to the LR scheduler.
        schedulers (sequence): Specifies the sequence of chained
            schedulers or their configurations.
        milestones (list): Specifies the list of integers that
            reflects milestone points.
        last_epoch (int, optional): Specifies the index of last epoch.
            Default: ``-1``

    Returns:
        ``SequentialLR``: The instantiated sequential LR scheduler.

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.lr_schedulers import create_sequential_lr
        >>> optimizer = torch.optim.SGD(torch.nn.Linear(4, 6).parameters(), lr=0.01)
        >>> scheduler = create_sequential_lr(
        ...     optimizer,
        ...     schedulers=[
        ...         {
        ...             "_target_": "torch.optim.lr_scheduler.LinearLR",
        ...             "start_factor": 0.001,
        ...             "end_factor": 1.0,
        ...             "total_iters": 2,
        ...         },
        ...         {
        ...             "_target_": "torch.optim.lr_scheduler.LinearLR",
        ...             "start_factor": 1.0,
        ...             "end_factor": 0.001,
        ...             "total_iters": 8,
        ...         },
        ...     ],
        ...     milestones=[2],
        ... )
        <torch.optim.lr_scheduler.SequentialLR at 0x012345678>
    """
    return SequentialLR(
        optimizer,
        schedulers=[setup_lr_scheduler(optimizer, scheduler) for scheduler in schedulers],
        milestones=milestones,
        last_epoch=last_epoch,
    )


def create_linear_warmup_cosine_decay_lr(
    optimizer: Optimizer,
    num_warmup_steps: int,
    num_total_steps: int,
    start_factor: float = 0.001,
    end_lr: float = 1e-6,
) -> SequentialLR:
    r"""Creates a LR scheduler with linear warm-up and cosine decay.

    Args:
        optimizer (``torch.optim.Optimizer``): Specifies the optimizer
            associated to the LR scheduler.
        num_warmup_steps (int): Specifies the number of steps for the
            warmup phase.
        num_total_steps (int): Specifies the number of total steps.
        start_factor (float, optional): Specifies the multiplication
            factor for the first step (``0``). Default: ``0.001``
        end_lr (float, optional): Specifies the learning rate for the
            last step (``num_total_steps``). Default: ``1e-6``

    Returns:
        ``torch.optim.lr_scheduler.SequentialLR``: The instantiated
            learning rate scheduler.
    """
    warmup = LinearLR(
        optimizer, start_factor=start_factor, end_factor=1.0, total_iters=num_warmup_steps
    )
    decay = CosineAnnealingLR(optimizer, T_max=num_total_steps - num_warmup_steps, eta_min=end_lr)
    return SequentialLR(optimizer, schedulers=[warmup, decay], milestones=[num_warmup_steps])


def create_linear_warmup_linear_decay_lr(
    optimizer: Optimizer,
    num_warmup_steps: int,
    num_total_steps: int,
    start_factor: float = 0.001,
    end_factor: float = 0.001,
) -> SequentialLR:
    r"""Creates a LR scheduler with linear warm-up and linear decay.

    Args:
        optimizer (``torch.optim.Optimizer``): Specifies the optimizer
            associated to the LR scheduler.
        num_warmup_steps (int): Specifies the number of steps for the
            warmup phase.
        num_total_steps (int): Specifies the number of total steps.
        start_factor (float, optional): Specifies the multiplication
            factor for the first step (``0``). Default: ``0.001``
        end_factor (float, optional): Specifies the multiplication
            factor for the last step (``num_total_steps``).
            Default: ``0.001``

    Returns:
        ``torch.optim.lr_scheduler.SequentialLR``: The instantiated
            learning rate scheduler.
    """
    warmup = LinearLR(
        optimizer, start_factor=start_factor, end_factor=1.0, total_iters=num_warmup_steps
    )
    decay = LinearLR(
        optimizer,
        start_factor=1.0,
        end_factor=end_factor,
        total_iters=num_total_steps - num_warmup_steps,
    )
    return SequentialLR(optimizer, schedulers=[warmup, decay], milestones=[num_warmup_steps])



__all__ = [
    "DictBatcher",
    "DictOfListConverter",
    "DirFilter",
    "FileFilter",
    "ListOfDictConverter",
    "Looper",
    "PathLister",
    "PickleSaver",
    "PyTorchSaver",
    "SourceWrapper",
    "TensorDictShuffler",
    "TupleBatcher",
    "setup_iter_datapipe",
]

from gravitorch.data.datapipes.iter.batching import (
    DictBatcherIterDataPipe as DictBatcher,
)
from gravitorch.data.datapipes.iter.batching import (
    TupleBatcherIterDataPipe as TupleBatcher,
)
from gravitorch.data.datapipes.iter.dictionary import (
    DictOfListConverterIterDataPipe as DictOfListConverter,
)
from gravitorch.data.datapipes.iter.dictionary import (
    ListOfDictConverterIterDataPipe as ListOfDictConverter,
)
from gravitorch.data.datapipes.iter.factory import setup_iter_datapipe
from gravitorch.data.datapipes.iter.length import LooperIterDataPipe as Looper
from gravitorch.data.datapipes.iter.path import DirFilterIterDataPipe as DirFilter
from gravitorch.data.datapipes.iter.path import FileFilterIterDataPipe as FileFilter
from gravitorch.data.datapipes.iter.path import PathListerIterDataPipe as PathLister
from gravitorch.data.datapipes.iter.saving import PickleSaverIterDataPipe as PickleSaver
from gravitorch.data.datapipes.iter.saving import (
    PyTorchSaverIterDataPipe as PyTorchSaver,
)
from gravitorch.data.datapipes.iter.shuffling import (
    TensorDictShufflerIterDataPipe as TensorDictShuffler,
)
from gravitorch.data.datapipes.iter.source import (
    SourceWrapperIterDataPipe as SourceWrapper,
)



__all__ = [
    "AdvancedCoreCreator",
    "BaseCoreCreator",
    "VanillaCoreCreator",
    "setup_core_creator",
]

from gravitorch.creators.core.advanced import AdvancedCoreCreator
from gravitorch.creators.core.base import BaseCoreCreator, setup_core_creator
from gravitorch.creators.core.vanilla import VanillaCoreCreator



"""Helper functions to help to reproduce experiments."""

__all__ = [
    "BaseRandomSeedSetter",
    "NumpyRandomSeedSetter",
    "RandomRandomSeedSetter",
    "RandomSeedSetter",
    "TorchRandomSeedSetter",
    "get_random_seed",
    "get_torch_generator",
    "manual_seed",
    "numpy_seed",
    "torch_seed",
]

import random
from abc import ABC, abstractmethod
from contextlib import contextmanager
from typing import Optional, Union

import numpy
import torch

from gravitorch.utils.format import str_add_indent, to_pretty_dict_str


def get_random_seed(seed: int) -> int:
    r"""Gets a random seed.

    Args:
        seed (int): Specifies a random seed to make the process
            reproducible.

    Returns:
        int: A random seed. The value is between ``-2 ** 63`` and
            ``2 ** 63 - 1``.

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.seed import get_random_seed
        >>> get_random_seed(44)
        6176747449835261347
    """
    return torch.randint(
        -(2**63), 2**63 - 1, size=(1,), generator=get_torch_generator(seed)
    ).item()


def get_torch_generator(
    random_seed: int = 1, device: Union[torch.device, str, None] = "cpu"
) -> torch.Generator:
    r"""Creates a ``torch.Generator`` initialized with a given seed.

    Args:
        random_seed (int, optional): Specifies a random seed.
            Default: ``1``
        device (``torch.device`` or str or ``None``, optional):
            Specifies the desired device for the generator.
            Default: ``'cpu'``

    Returns:
        ``torch.Generator``

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.utils.seed import get_torch_generator
        >>> generator = get_torch_generator(42)
        >>> torch.rand(2, 4, generator=generator)
        tensor([[0.8823, 0.9150, 0.3829, 0.9593],
                [0.3904, 0.6009, 0.2566, 0.7936]])
        >>> generator = get_torch_generator(42)
        >>> torch.rand(2, 4, generator=generator)
        tensor([[0.8823, 0.9150, 0.3829, 0.9593],
                [0.3904, 0.6009, 0.2566, 0.7936]])
    """
    generator = torch.Generator(device)
    generator.manual_seed(random_seed)
    return generator


class BaseRandomSeedSetter(ABC):
    r"""Define the base class to implement a random seed setter.

    Each child class must implement the method ``manual_seed``.
    """

    @abstractmethod
    def manual_seed(self, seed: int) -> None:
        r"""Sets the seed for generating random numbers.

        Args:
            seed (int): Specifies the desired seed.

        Example usage:

        .. code-block:: python

            >>> from gravitorch.utils.seed import BaseRandomSeedSetter
            >>> setter: BaseRandomSeedSetter = ...  # Initialize a random seed setter.
            >>> setter.manual_seed(42)
        """


class NumpyRandomSeedSetter(BaseRandomSeedSetter):
    r"""Implements a random seed setter for the library ``numpy``.

    The seed must be between ``0`` and ``2**32 - 1``, so a modulo
    operator to convert an integer to an integer between ``0`` and
    ``2**32 - 1``.
    """

    def __str__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def manual_seed(self, seed: int) -> None:
        numpy.random.seed(seed % 2**32)


class RandomRandomSeedSetter(BaseRandomSeedSetter):
    r"""Implements a random seed setter for the python standard library
    ``random``."""

    def __str__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def manual_seed(self, seed: int) -> None:
        random.seed(seed)


class TorchRandomSeedSetter(BaseRandomSeedSetter):
    r"""Implements a random seed setter for the library ``torch``."""

    def __str__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def manual_seed(self, seed: int) -> None:
        torch.manual_seed(seed)
        if torch.cuda.is_available():
            torch.cuda.manual_seed_all(seed)


class RandomSeedSetter(BaseRandomSeedSetter):
    r"""Implements the default random seed setter.

    By default, it is initialized with the following random seed setters:

        - ``'numpy'``: ``NumpyRandomSeedSetter``
        - ``'random'``: ``RandomRandomSeedSetter``
        - ``'torch'``: ``TorchRandomSeedSetter``
    """

    registry: dict[str, BaseRandomSeedSetter] = {
        "numpy": NumpyRandomSeedSetter(),
        "random": RandomRandomSeedSetter(),
        "torch": TorchRandomSeedSetter(),
    }

    def __repr__(self) -> str:
        setters = {key: value for key, value in self.registry.items()}
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  {str_add_indent(to_pretty_dict_str(setters))}\n)"
        )

    def manual_seed(self, seed: int) -> None:
        for value in self.registry.values():
            value.manual_seed(seed)

    @classmethod
    def add_setter(cls, name: str, setter: "BaseRandomSeedSetter", exist_ok: bool = False) -> None:
        r"""Adds a random seed setter for a given name.

        Args:
            name (str): Specifies the name for the setter.
            setter (``BaseRandomSeedSetter``): Specifies the random
                seed setter to add.
            exist_ok (bool, optional): If ``False``, ``ValueError`` is
                raised if the name already exists. This parameter
                should be set to ``True`` to overwrite the setter for
                a name. Default: ``False``.

        Raises:
            ValueError if a random seed setter is already registered
                for the name and ``exist_ok=False``.

        Example usage:

        .. code-block:: python

            >>> from gravitorch.utils.seed import BaseRandomSeedSetter, RandomSeedSetter
            >>> class OtherRandomSeedSetter(BaseRandomSeedSetter):
            ...    def manual_seed(self, seed: int) -> None:
            ...        ...  # Custom implementation
            >>> RandomSeedSetter.add_setter('other', OtherRandomSeedSetter())
            # To overwrite an existing random seed setter
            >>> RandomSeedSetter.add_setter('other', OtherRandomSeedSetter(), exist_ok=True)
        """
        if name in cls.registry and not exist_ok:
            raise ValueError(
                f"A setter ({cls.registry[name]}) is already registered for the name {name}. "
                "Please use `exist_ok=True` if you want to overwrite the setter for this name"
            )
        cls.registry[name] = setter

    @classmethod
    def has_setter(cls, name: str) -> bool:
        r"""Indicates if a random seed setter is registered for the given name.

        Args:
            name (str): Specifies the name to check.

        Returns:
            bool: ``True`` if a random seed setter is registered,
                otherwise ``False``.

        Example usage:

        .. code-block:: python

            >>> from gravitorch.utils.seed import RandomSeedSetter
            >>> RandomSeedSetter.has_setter('torch')
            True
            >>> RandomSeedSetter.has_setter('missing')
            False
        """
        return name in cls.registry


def manual_seed(seed: int, setter: Optional[BaseRandomSeedSetter] = None) -> None:
    r"""Sets the seed for generating random numbers.

    Args:
        seed (int): Specifies the desired seed.
        setter (``BaseRandomSeedSetter`` or ``None``, optional):
            Specifies the random seed setters. If ``None``, the
            default random seed setter (``RandomSeedSetter``) is used.
            Default: ``None``

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.seed import manual_seed
        >>> manual_seed(42)
        >>> torch.randn(3)
        tensor([0.3367, 0.1288, 0.2345])
        >>> torch.randn(3)
        tensor([ 0.2303, -1.1229, -0.1863])
        >>> manual_seed(42)
        >>> torch.randn(3)
        tensor([0.3367, 0.1288, 0.2345])

        # Set the seed only for numpy
        >>> from gravitorch.utils.seed import NumpyRandomSeedSetter
        >>> manual_seed(42, NumpyRandomSeedSetter())
        >>> torch.randn(3)
        tensor([ 2.2082, -0.6380,  0.4617])
    """
    setter = setter or RandomSeedSetter()
    setter.manual_seed(seed)


@contextmanager
def numpy_seed(seed: int):
    r"""Implements a context manager to manage the NumPy random seed and random
    number generator (RNG) state.

    The context manager sets the specified NumPy random seed and
    restores the original RNG state afterward.

    Args:
        seed (int): Specifies the random number generator seed to use
            while using this context manager.

    Example usage:

    .. code-block:: python

        >>> import numpy
        >>> from gravitorch.utils.seed import numpy_seed
        >>> with numpy_seed(42):
        ...     print(numpy.random.randn(2, 4))
        [[ 0.49671415 -0.1382643   0.64768854  1.52302986]
         [-0.23415337 -0.23413696  1.57921282  0.76743473]]
        >>> with numpy_seed(42):
        ...     print(numpy.random.randn((2, 4)))
        [[ 0.49671415 -0.1382643   0.64768854  1.52302986]
         [-0.23415337 -0.23413696  1.57921282  0.76743473]]
    """
    state = numpy.random.get_state()
    try:
        NumpyRandomSeedSetter().manual_seed(seed)
        yield
    finally:
        numpy.random.set_state(state)


@contextmanager
def torch_seed(seed: int):
    r"""Implements a context manager to manage the PyTorch random seed and
    random number generator (RNG) state.

    The context manager sets the specified PyTorch random seed and
    restores the original RNG state afterward.

    Args:
        seed (int): Specifies the random number generator seed to use
            while using this context manager.

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.utils.seed import torch_seed
        >>> with torch_seed(42):
        ...     print(torch.randn(2, 4))
        tensor([[ 0.3367,  0.1288,  0.2345,  0.2303],
                [-1.1229, -0.1863,  2.2082, -0.6380]])
        >>> with torch_seed(42):
        ...     print(torch.randn(2, 4))
        tensor([[ 0.3367,  0.1288,  0.2345,  0.2303],
                [-1.1229, -0.1863,  2.2082, -0.6380]])
    """
    state = torch.get_rng_state()
    cuda_states = torch.cuda.get_rng_state_all()
    try:
        TorchRandomSeedSetter().manual_seed(seed)
        yield
    finally:
        torch.set_rng_state(state)
        torch.cuda.set_rng_state_all(cuda_states)



__all__ = ["NoOpProfiler"]

import logging

from gravitorch.utils.profilers.base import BaseProfiler

logger = logging.getLogger(__name__)


class NoOpProfiler(BaseProfiler):
    r"""Implements a no-op profiler.

    This class allows to use a profiler without changing the
    implementation.
    """

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def __str__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def step(self) -> None:
        r"""Signals the profiler that the next profiling step has started.

        Because it is a no-op class, this method does nothing.
        """



__all__ = ["ExponentialMovingAverage", "MovingAverage"]

from collections import deque
from collections.abc import Iterable
from typing import Any, Union

import torch

from gravitorch.utils.meters.exceptions import EmptyMeterError


class MovingAverage:
    r"""Defines a class to compute and store the exponential moving average
    value of float number.

    Args:
        values (iterable, optional): Specifies the initial values.
            Default: ``tuple()``
        window_size (int, optional): Specifies the maximum window
            size. Default: ``20``
    """

    def __init__(self, values: Iterable[Union[int, float]] = (), window_size: int = 20):
        self._deque = deque(values, maxlen=window_size)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(window_size={self.window_size:,})"

    @property
    def values(self) -> tuple[Union[int, float]]:
        return tuple(self._deque)

    @property
    def window_size(self) -> int:
        return self._deque.maxlen

    def clone(self) -> "MovingAverage":
        r"""Creates a copy of the current meter.

        Returns:
            ``MovingAverage``: A copy of the current meter.
        """
        return MovingAverage(values=tuple(self._deque), window_size=self.window_size)

    def equal(self, other) -> bool:
        r"""Indicates if two meters are equal or not.

        Args:
            other: Specifies the value to compare.

        Returns:
            bool: ``True`` if the meters are equal,
                ``False`` otherwise.
        """
        if not isinstance(other, MovingAverage):
            return False
        return self.state_dict() == other.state_dict()

    def load_state_dict(self, state_dict: dict[str, Any]) -> None:
        r"""Loads a state to the history tracker.

        Args:
            state_dict (dict): Specifies a dictionary containing state
                keys with values.
        """
        self._deque = deque(state_dict["values"], maxlen=state_dict["window_size"])

    def reset(self) -> None:
        r"""Reset the meter."""
        self._deque.clear()

    def smoothed_average(self) -> float:
        r"""Computes the smoothed average value.

        Returns:
            float: The smoothed average value.

        Raises:
            ``EmptyMeterError`` if the meter is empty.
        """
        if not self._deque:
            raise EmptyMeterError("The moving average meter is empty")
        return torch.tensor(self.values).float().mean().item()

    def state_dict(self) -> dict[str, Any]:
        r"""Returns a dictionary containing state values.

        Returns:
            dict: The state values in a dict.
        """
        return {"values": self.values, "window_size": self.window_size}

    def update(self, value: Union[int, float]) -> None:
        r"""Updates the meter given a new value.

        Args:
            value (int or float): Specifies the value to add to the
                meter.
        """
        self._deque.append(value)


class ExponentialMovingAverage:
    r"""Defines a class to compute and store the exponential moving average
    value of float number.

    Args:
        alpha (float, optional): Specifies the smoothing factor such
            as ``0 < alpha < 1``.
        count (int, optional): Specifies the initial count value.
            Default: ``0``
        smoothed_average (float, optional): Specifies the initial
            value of the smoothed average. Default: ``0.0``
    """

    def __init__(
        self,
        alpha: float = 0.98,
        count: int = 0,
        smoothed_average: float = 0.0,
    ):
        self._alpha = float(alpha)
        self._count = int(count)
        self._smoothed_average = float(smoothed_average)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(alpha={self._alpha}, count={self._count:,}, "
            f"smoothed_average={self._smoothed_average}, )"
        )

    @property
    def count(self) -> int:
        r"""int: The number of examples in the meter since the last
        reset."""
        return self._count

    def clone(self) -> "ExponentialMovingAverage":
        r"""Creates a copy of the current meter.

        Returns:
            ``ExponentialMovingAverage``: A copy of the current meter.
        """
        return ExponentialMovingAverage(
            alpha=self._alpha,
            count=self._count,
            smoothed_average=self._smoothed_average,
        )

    def equal(self, other) -> bool:
        r"""Indicates if two meters are equal or not.

        Args:
            other: Specifies the value to compare.

        Returns:
            bool: ``True`` if the meters are equal,
                ``False`` otherwise.
        """
        if not isinstance(other, ExponentialMovingAverage):
            return False
        return self.state_dict() == other.state_dict()

    def load_state_dict(self, state_dict: dict[str, Any]) -> None:
        r"""Loads a state to the history tracker.

        Args:
            state_dict (dict): Specifies a dictionary containing state
                keys with values.
        """
        self._alpha = state_dict["alpha"]
        self._count = state_dict["count"]
        self._smoothed_average = state_dict["smoothed_average"]

    def reset(self) -> None:
        r"""Reset the meter."""
        self._count = 0
        self._smoothed_average = 0.0

    def smoothed_average(self) -> float:
        r"""Computes the smoothed average value.

        Returns:
            float: The smoothed average value.

        Raises:
            ``EmptyMeterError`` if the meter is empty.
        """
        if not self._count:
            raise EmptyMeterError("The exponential moving average meter is empty")
        return self._smoothed_average

    def state_dict(self) -> dict[str, Any]:
        r"""Returns a dictionary containing state values.

        Returns:
            dict: The state values in a dict.
        """
        return {
            "alpha": self._alpha,
            "count": self._count,
            "smoothed_average": self._smoothed_average,
        }

    def update(self, value: float) -> None:
        r"""Updates the meter given a new value.

        Args:
            value (float): Specifies the value to add to the meter.
        """
        self._smoothed_average = self._alpha * self._smoothed_average + (1.0 - self._alpha) * value
        self._count += 1



from __future__ import annotations

__all__ = [
    "BATCH_LOAD_TIME_AVG_MS",
    "BATCH_LOAD_TIME_MAX_MS",
    "BATCH_LOAD_TIME_MEDIAN_MS",
    "BATCH_LOAD_TIME_MIN_MS",
    "BATCH_LOAD_TIME_PCT",
    "BATCH_LOAD_TIME_STDDEV_MS",
    "BatchLoadingTimer",
    "EPOCH_TIME_SEC",
    "ITER_TIME_AVG_MS",
    "NUM_BATCHES",
    "sync_perf_counter",
    "timeblock",
]

import logging
import time
from collections.abc import Generator, Iterable
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any, TypeVar

import torch

from gravitorch.utils.exp_trackers.steps import EpochStep
from gravitorch.utils.format import convert_seconds_to_readable_format
from gravitorch.utils.meters import ScalarMeter

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine

logger = logging.getLogger(__name__)

T = TypeVar("T")

BATCH_LOAD_TIME_AVG_MS = "batch_load_time_avg_ms"
BATCH_LOAD_TIME_MAX_MS = "batch_load_time_max_ms"
BATCH_LOAD_TIME_MEDIAN_MS = "batch_load_time_median_ms"
BATCH_LOAD_TIME_MIN_MS = "batch_load_time_min_ms"
BATCH_LOAD_TIME_PCT = "batch_load_time_pct"
BATCH_LOAD_TIME_STDDEV_MS = "batch_load_time_stddev_ms"
EPOCH_TIME_SEC = "epoch_time_sec"
ITER_TIME_AVG_MS = "iter_time_avg_ms"
NUM_BATCHES = "num_batches"


def sync_perf_counter() -> float:
    r"""Extension of ``time.perf_counter`` that waits for all kernels in all
    streams on a CUDA device to complete.

    Returns:
        float: Same as ``time.perf_counter()``.
            See https://docs.python.org/3/library/time.html#time.perf_counter
            for more information.

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.timing import sync_perf_counter
        >>> tic = sync_perf_counter()
        >>> x = [1, 2, 3]
        >>> toc = sync_perf_counter()
        >>> toc - tic
        7.040074708999999
    """
    if torch.cuda.is_available():
        torch.cuda.synchronize()
    return time.perf_counter()


@contextmanager
def timeblock(message: str = "Total time: {time}") -> Generator[None, None, None]:
    r"""Implements a context manager to measure the execution time of a block of
    code.

    Args:
        message (str, optional): Specifies the message displayed when
            the time is logged.

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.timing import timeblock
        >>> with timeblock():
        ...     x = [1, 2, 3]
        INFO:gravitorch.utils.time_tracking:Total time: 0:00:00.000039
        >>> with timeblock("Training: {time}"):
        ...     x = [1, 2, 3]
        INFO:gravitorch.utils.time_tracking:Training: 0:00:00.000035
    """
    if "{time}" not in message:
        raise ValueError(f"{time} is missing in the message (received: {message})")
    start_time = sync_perf_counter()
    try:
        yield
    finally:
        logger.info(
            message.format(
                time=convert_seconds_to_readable_format(sync_perf_counter() - start_time)
            )
        )


class BatchLoadingTimer(Iterable[T]):
    r"""Implements an iterator around a batch loader iterable to monitor the
    time performances.

    Args:
        batch_loader (Iterable): Specifies the batch loader to measure
            the time performances.
        epoch (int): Specifies the epoch.
        prefix (str): Specifies the prefix used to log the values.


    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.timing import BatchLoadingTimer
        >>> my_batch_loader = [1, 2, 3, 4, 5]
        >>> batch_loader = BatchLoadingTimer(my_batch_loader, epoch=0, prefix="train")
        >>> for batch in batch_loader:
        ...     pass # do something
        >>> batch_loader.get_stats()
        {'batch_load_time_avg_ms': 0.003508200001078876,
         'batch_load_time_max_ms': 0.010916999997334642,
         'batch_load_time_median_ms': 0.0016660000028423383,
         'batch_load_time_min_ms': 0.0015419999996879596,
         'batch_load_time_pct': 36.012564686438395,
         'batch_load_time_stddev_ms': 0.004143146270507714,
         'iter_time_avg_ms': 0.00974159999884705,
         'epoch_time_sec': 4.870799999423525e-05,
         'num_batches': 5}
        >>> batch_loader.log_stats()
        INFO:gravitorch.utils.timing:Epoch: 0 | batch: 5 | total time: 0.000 s | avg iter time: 0.010 ms | avg batch load time: 0.004 ms | avg batch load pct: 36.01 %  # noqa: E501,B950
        INFO:gravitorch.utils.timing:Batch loading time min/avg/median/max/stddev = 0.002/0.004/0.002/0.011/0.004 ms  # noqa: E501,B950
    """

    def __init__(self, batch_loader: Iterable[T], epoch: int, prefix: str):
        self._batch_loader = batch_loader
        self._epoch = epoch
        self._prefix = prefix
        self._meter = ScalarMeter(max_size=100)

        self._start_time: float | None = None
        self._end_time: float | None = None

    def __iter__(self):
        self._start_time = sync_perf_counter()
        tic = self._start_time
        for batch in self._batch_loader:
            self._meter.update(sync_perf_counter() - tic)
            yield batch
            tic = sync_perf_counter()
        self._end_time = sync_perf_counter()

    def get_stats(self) -> dict[str, Any]:
        r"""Gets some stats about the batch loading performances.

        The computed stats are:

            - ``'batch_load_time_avg_ms'``: The average time in
                milliseconds to load a batch.
            - ``'batch_load_time_max_ms'``: The maximum time in
                milliseconds to load a batch.
            - ``'batch_load_time_median_ms'``: The median time in
                milliseconds to load a batch.
            - ``'batch_load_time_min_ms'``: The minimum time in
                milliseconds to load a batch.
            - ``'batch_load_time_stddev_ms'``: The standard deviation
                time in milliseconds to load a batch.
            - ``'num_batches'``: The number of batches.
            - ``'epoch_time_sec'``: The total time in seconds to
                iterate and process all the batches. It corresponds
                to the time to computed one epoch.
            - ``'iter_time_avg_ms'``: The average time in
                milliseconds to iterate and process a batch. It
                corresponds to the time to load and process a batch.
            - ``'batch_load_time_pct'``: The percentage of time spent
                to load the batches. The lower the better. This metric
                is important to know if the batch loading is a
                bottleneck or not.

        This method has to be called after the batches have been
        iterated.

        Returns:
            dict: The dictionary with the batch loading stats. The
                dictionary is empty if there is no batches.
        """
        if self._meter.count == 0:
            return {}
        epoch_time_sec = self._end_time - self._start_time
        iter_time_avg_ms = 1000 * epoch_time_sec / self._meter.count
        batch_load_avg_ms = 1000 * self._meter.average()
        return {
            BATCH_LOAD_TIME_AVG_MS: batch_load_avg_ms,
            BATCH_LOAD_TIME_MAX_MS: 1000 * self._meter.max(),
            BATCH_LOAD_TIME_MEDIAN_MS: 1000 * self._meter.median(),
            BATCH_LOAD_TIME_MIN_MS: 1000 * self._meter.min(),
            BATCH_LOAD_TIME_PCT: 100 * batch_load_avg_ms / iter_time_avg_ms,
            BATCH_LOAD_TIME_STDDEV_MS: 1000 * self._meter.std(),
            ITER_TIME_AVG_MS: iter_time_avg_ms,
            EPOCH_TIME_SEC: epoch_time_sec,
            NUM_BATCHES: int(self._meter.count),
        }

    def log_stats(self, engine: BaseEngine | None = None) -> None:
        r"""Logs the time statistics about the batch loading.

        This method has to be called after the batches have been
        iterated.

        Args:
            engine (``BaseEngine`` or None): Specifies an engine if
                you want to log the time metrics in the engine. If
                ``None``, the time metrics are not logged in an
                engine. In each case, the time metrics are logged in
                the python logger. Default: ``None``
        """
        if self._meter.count == 0:
            return  # Do nothing if there is no mini-batch.
        stats = self.get_stats()
        epoch_time_sec = stats[EPOCH_TIME_SEC]
        epoch_time_str = (
            convert_seconds_to_readable_format(epoch_time_sec)
            if epoch_time_sec > 10
            else f"{epoch_time_sec:.3f} s"
        )
        logger.info(
            f"Epoch: {self._epoch:,} | batch: {stats[NUM_BATCHES]:,} | "
            f"total time: {epoch_time_str} | "
            f"avg iter time: {stats[ITER_TIME_AVG_MS]:.3f} ms | "
            f"avg batch load time: {stats[BATCH_LOAD_TIME_AVG_MS]:,.3f} ms | "
            f"avg batch load pct: {stats[BATCH_LOAD_TIME_PCT]:.2f} %"
        )
        logger.info(
            f"Batch loading time min/avg/median/max/stddev = {stats[BATCH_LOAD_TIME_MIN_MS]:.3f}/"
            f"{stats[BATCH_LOAD_TIME_AVG_MS]:.3f}/{stats[BATCH_LOAD_TIME_MEDIAN_MS]:.3f}/"
            f"{stats[BATCH_LOAD_TIME_MAX_MS]:.3f}/{stats[BATCH_LOAD_TIME_STDDEV_MS]:.3f} ms"
        )
        if engine:
            engine.log_metrics(
                {f"{self._prefix}{key}": value for key, value in stats.items()},
                step=EpochStep(self._epoch),
            )



__all__ = [
    "BaseEngineSaver",
    "BestEngineStateSaver",
    "BestHistorySaver",
    "EpochEngineStateSaver",
    "LastHistorySaver",
    "TagEngineStateSaver",
]

import logging
from abc import abstractmethod
from pathlib import Path
from typing import Union

from gravitorch.distributed import comm as dist
from gravitorch.engines.base import BaseEngine
from gravitorch.engines.events import EngineEvents
from gravitorch.handlers.base import BaseHandler
from gravitorch.handlers.utils import add_unique_event_handler
from gravitorch.utils.events import VanillaEventHandler
from gravitorch.utils.history import get_best_values, get_last_values
from gravitorch.utils.io import save_pytorch
from gravitorch.utils.path import sanitize_path

logger = logging.getLogger(__name__)


class BaseEngineSaver(BaseHandler):
    r"""Implements a base handler to save something about the engine.

    Note that it is not the only way to save something about the
    engine. Feel free to use another approaches if this approach does
    not fit your needs.

    The child class has to implement the ``_save`` method.

    Args:
        path (``pathlib.Path`` or str): Specifies the path to the
            folder where to save the artifacts.
        event (str): Specifies the event used to save the artifacts.
        only_main_process (bool, optional): If ``True``, only the
            main process saves the artifacts, otherwise
            all the processes save the artifacts.
            Default: ``True``
    """

    def __init__(
        self,
        path: Union[Path, str],
        event: str,
        only_main_process: bool = True,
    ):
        self._path = sanitize_path(path)
        self._event = str(event)
        self._only_main_process = only_main_process

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  event={self._event},\n"
            f"  path={self._path},\n"
            f"  only_main_process={self._only_main_process},\n"
            ")"
        )

    def attach(self, engine: BaseEngine) -> None:
        add_unique_event_handler(
            engine=engine,
            event=self._event,
            event_handler=VanillaEventHandler(self.save, handler_kwargs={"engine": engine}),
        )

    def save(self, engine: BaseEngine) -> None:
        r"""Saves the values associated to the histories.

        Args:
            engine (``BaseEngine``): Specifies the engine.
        """
        if self._only_main_process and not dist.is_main_process():
            return
        self._save(engine)

    @abstractmethod
    def _save(self, engine: BaseEngine) -> None:
        r"""Saves the values associated to the histories.

        Args:
            engine (``BaseEngine``): Specifies the engine.
        """


###################
#     History     #
###################


class BestHistorySaver(BaseEngineSaver):
    r"""Implements a handler to save the best history values in a PyTorch
    file."""

    def __init__(
        self,
        path: Union[Path, str],
        event: str = EngineEvents.COMPLETED,
        only_main_process: bool = True,
    ):
        super().__init__(path=path, event=event, only_main_process=only_main_process)

    def _save(self, engine: BaseEngine) -> None:
        path = self._path.joinpath("history_best.pt")
        logger.info(f"Saving best history values in {path}")
        save_pytorch(get_best_values(engine.get_histories()), path)


class LastHistorySaver(BaseEngineSaver):
    r"""Implements a handler to save the last history values in a PyTorch
    file."""

    def __init__(
        self,
        path: Union[Path, str],
        event: str = EngineEvents.EPOCH_COMPLETED,
        only_main_process: bool = True,
    ):
        super().__init__(path=path, event=event, only_main_process=only_main_process)

    def _save(self, engine: BaseEngine) -> None:
        path = self._path.joinpath("history_last.pt")
        logger.info(f"Saving last history values in {path}")
        save_pytorch(get_last_values(engine.get_histories()), path)


#################
#     state     #
#################


class BestEngineStateSaver(BaseEngineSaver):
    r"""Implements a handler to save the engine state dict in a PyTorch file.

    This handler saves the best checkpoint of a set of metrics.
    Internally, this handler uses the history to know if a
    metric has improved or not.

    Note it is usually recommended to save the state dict after the
    other handlers were run to capture the correct engine state.

    Args:
        path (``pathlib.Path`` or str): Specifies the path to the
            folder where to write the state dict.
        keys (tuple or list): Specifies the set of metrics to create
            the best checkpoint. Each key should be associated to a
            comparable history in the engine.
        event (str, optional): Specifies the event used to save the
            engine state dict. Default: ``'epoch_completed'``
        only_main_process (bool, optional): If ``True``, only the main
            process saves the engine state dict, otherwise all the
            processes save the engine state dict. Default: ``True``
    """

    def __init__(
        self,
        path: Union[Path, str],
        keys: Union[tuple[str, ...], list[str]],
        event: str = EngineEvents.EPOCH_COMPLETED,
        only_main_process: bool = True,
    ):
        super().__init__(path=path, event=event, only_main_process=only_main_process)
        self._keys = tuple(keys)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  event={self._event},\n"
            f"  path={self._path},\n"
            f"  keys={self._keys},\n"
            f"  only_main_process={self._only_main_process},\n"
            ")"
        )

    def _save(self, engine: BaseEngine) -> None:
        r"""Saves the engine state dict in a PyTorch file.

        Args:
            engine (``BaseEngine``): Specifies the engine.
        """
        logger.info("Saving the best checkpoints according to some metrics...")
        for key in self._keys:
            if not engine.has_history(key):
                logger.warning(
                    f"Attempting to save the best state dict for '{key}' but there is no "
                    "history so this metric is skipped"
                )
                continue
            history = engine.get_history(key)
            if not history.is_comparable():
                logger.warning(
                    f"Attempting to save the best state dict for '{key}' but the history "
                    "is not comparable so this metric is skipped"
                )
                continue
            if history.is_empty():
                logger.warning(
                    f"Attempting to save the best state dict for '{key}' but the history "
                    "is empty so this metric is skipped"
                )
                continue
            if history.has_improved():
                logger.info(f"Saving 'best_{key}' engine state dict")
                save_pytorch(
                    engine.state_dict(),
                    self._path.joinpath(f"best_{key}/ckpt_engine_{dist.get_rank()}.pt"),
                )


class EpochEngineStateSaver(BaseEngineSaver):
    r"""Implements a handler to save the engine state dict in a PyTorch file.

    This engine state dict saver creates a new file for each epoch.

    Note it is usually recommended to save the state dict after the
    other handlers were run to capture the correct engine state.
    """

    def __init__(
        self,
        path: Union[Path, str],
        event: str = EngineEvents.EPOCH_COMPLETED,
        only_main_process: bool = True,
    ):
        super().__init__(path=path, event=event, only_main_process=only_main_process)

    def _save(self, engine: BaseEngine) -> None:
        r"""Saves the engine state dict in a PyTorch file.

        Args:
            engine (``BaseEngine``): Specifies the engine.
        """
        logger.info(f"Saving 'epoch {engine.epoch}' engine state dict")
        save_pytorch(
            engine.state_dict(),
            self._path.joinpath(f"epoch/ckpt_engine_{engine.epoch}_{dist.get_rank()}.pt"),
        )


class TagEngineStateSaver(BaseEngineSaver):
    r"""Implements a handler to save the engine state dict in a PyTorch file.

    This handler overrides the previous file everytime the ``save``
    method is called.

    Note it is usually recommended to save the state dict after the
    other handlers were run to capture the correct engine state.

    Args:
        path (``pathlib.Path`` or str): Specifies the path to the
            folder where to save the state dict.
        event (str, optional): Specifies the event used to save the
            engine state dict. Default: ``'epoch_completed'``
        tag (str, optional): Specifies the tag used to create the
            checkpoint name. Default: ``'last'``
        only_main_process (bool, optional): If ``True``, only the main
            process saves the engine state dict, otherwise all the
            processes save the engine state dict. Default: ``True``
    """

    def __init__(
        self,
        path: Union[Path, str],
        event: str = EngineEvents.EPOCH_COMPLETED,
        tag: str = "last",
        only_main_process: bool = True,
    ):
        super().__init__(path=path, event=event, only_main_process=only_main_process)
        self._tag = str(tag)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  event={self._event},\n"
            f"  path={self._path},\n"
            f"  tag={self._tag},\n"
            f"  only_main_process={self._only_main_process},\n"
            ")"
        )

    def _save(self, engine: BaseEngine) -> None:
        logger.info(f"Saving '{self._tag}' engine state dict...")
        save_pytorch(
            engine.state_dict(),
            self._path.joinpath(f"{self._tag}/ckpt_engine_{dist.get_rank()}.pt"),
        )



__all__ = ["is_loss_decreasing", "is_loss_decreasing_with_adam", "is_loss_decreasing_with_sgd"]

import logging
from collections.abc import Callable
from typing import Union

from torch import Tensor
from torch.nn import Module
from torch.nn.utils import clip_grad_norm_
from torch.optim import SGD, Adam, Optimizer

from gravitorch.nn.utils.module_helpers import module_mode
from gravitorch.utils.seed import torch_seed

logger = logging.getLogger(__name__)


def is_loss_decreasing(
    module: Module,
    criterion: Union[Module, Callable[[Tensor, Tensor], Tensor]],
    optimizer: Optimizer,
    feature: Tensor,
    target: Tensor,
    num_iterations: int = 1,
    random_seed: int = 10772155803920552556,
) -> bool:
    r"""Checks if the loss decreased after some iterations.

    Args:
        module (``torch.nn.Module``): Specifies the module to test.
        criterion (``torch.nn.Module`` or callable): Specifies the
            criterion.
        optimizer (``torch.optim.Optimizer``): Specifies the optimizer
            to update the weights of the model.
        feature (``torch.Tensor``): Specifies the input of the module.
        target (``torch.Tensor``): Specifies the target used to
            compute the loss.
        num_iterations (int, optional): Specifies the number of
            optimization steps. Default: ``1``
        random_seed (int, optional): Specifies the random seed.
            The random seed is used to have a deterministic when the
            loss value is computed. Default: ``10772155803920552556``

    Returns:
        bool: ``True`` if the loss decreased after some iterations,
            otherwise ``False``.

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from torch import nn
        >>> from gravitorch.nn.utils import is_loss_decreasing
        >>> module = nn.Linear(4, 2)
        >>> is_loss_decreasing(
        ...     module=module,
        ...     criterion=nn.MSELoss(),
        ...     optimizer=SGD(module.parameters(), lr=0.01),
        ...     feature=torch.rand(4, 4),
        ...     target=torch.rand(4, 2),
        ... )
        True
    """
    with module_mode(module):
        module.eval()
        with torch_seed(random_seed):
            initial_loss = criterion(module(feature), target).item()
        logger.debug(f"Initial loss value: {initial_loss}")

        module.train()
        for _ in range(num_iterations):
            optimizer.zero_grad()
            loss = criterion(module(feature), target)
            loss.backward()
            clip_grad_norm_(module.parameters(), max_norm=1, norm_type=2)
            optimizer.step()

        module.eval()
        with torch_seed(random_seed):
            final_loss = criterion(module(feature), target).item()
        logger.debug(f"Final loss value: {final_loss}")
    return final_loss < initial_loss


def is_loss_decreasing_with_adam(
    module: Module,
    criterion: Union[Module, Callable[[Tensor, Tensor], Tensor]],
    feature: Tensor,
    target: Tensor,
    lr: float = 0.0003,
    num_iterations: int = 1,
    random_seed: int = 10772155803920552556,
) -> bool:
    r"""Checks if the loss decreased after some iterations.

    The module is trained with the Adam optimizer.

    Args:
        module (``torch.nn.Module``): Specifies the module to test.
        criterion (``torch.nn.Module`` or callable): Specifies the
            criterion.
        feature (``torch.Tensor``): Specifies the input of the module.
        target (``torch.Tensor``): Specifies the target used to
            compute the loss.
        lr (float, optional): Specifies the learning rate.
            Default: ``0.0003``
        num_iterations (int, optional): Specifies the number of
            optimization steps. Default: ``1``
        random_seed (int, optional): Specifies the random seed.
            The random seed is used to have a deterministic when
            the loss value is computed.
            Default: ``10772155803920552556``

    Returns:
        bool: ``True`` if the loss decreased after some iterations,
            otherwise ``False``.

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from torch import nn
        >>> from gravitorch.nn.utils import is_loss_decreasing_with_adam
        >>> is_loss_decreasing_with_adam(
        ...     module=nn.Linear(4, 2),
        ...     criterion=nn.MSELoss(),
        ...     feature=torch.rand(4, 4),
        ...     target=torch.rand(4, 2),
        ...     lr=0.0003,
        ... )
        True
    """
    return is_loss_decreasing(
        module=module,
        criterion=criterion,
        optimizer=Adam(module.parameters(), lr=lr),
        feature=feature,
        target=target,
        num_iterations=num_iterations,
        random_seed=random_seed,
    )


def is_loss_decreasing_with_sgd(
    module: Module,
    criterion: Union[Module, Callable[[Tensor, Tensor], Tensor]],
    feature: Tensor,
    target: Tensor,
    lr: float = 0.01,
    num_iterations: int = 1,
    random_seed: int = 10772155803920552556,
) -> bool:
    r"""Checks if the loss decreased after some iterations.

    The module is trained with the ``torch.optim.SGD`` optimizer.

    Args:
        module (``torch.nn.Module``): Specifies the module to test.
        criterion (``torch.nn.Module`` or callable): Specifies the
            criterion.
        feature (``torch.Tensor``): Specifies the input of the module.
        target (``torch.Tensor``): Specifies the target used to
            compute the loss.
        lr (float, optional): Specifies the learning rate.
            Default: ``0.01``
        num_iterations (int, optional): Specifies the number of
            optimization steps. Default: ``1``
        random_seed (int, optional): Specifies the random seed. The
        random seed is used to have a deterministic when the loss
            value is computed. Default: ``10772155803920552556``

    Returns:
        bool: ``True`` if the loss decreased after some iterations,
            otherwise ``False``.

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from torch import nn
        >>> from gravitorch.nn.utils import is_loss_decreasing_with_adam
        >>> is_loss_decreasing_with_adam(
        ...     module=nn.Linear(4, 2),
        ...     criterion=nn.MSELoss(),
        ...     feature=torch.rand(4, 4),
        ...     target=torch.rand(4, 2),
        ...     lr=0.01,
        ... )
        True
    """
    return is_loss_decreasing(
        module=module,
        criterion=criterion,
        optimizer=SGD(module.parameters(), lr=lr),
        feature=feature,
        target=target,
        num_iterations=num_iterations,
        random_seed=random_seed,
    )



r"""This package contains the network base class and some implemented
networks."""

__all__ = [
    "AlphaMLP",
    "BaseMLP",
    "BetaMLP",
    "ImageClassificationNetwork",
    "LeNet5",
    "PyTorchMnistNet",
    "create_alpha_mlp",
    "create_beta_mlp",
]

from gravitorch.models.networks.image_classification import ImageClassificationNetwork
from gravitorch.models.networks.lenet import LeNet5
from gravitorch.models.networks.mlp import (
    AlphaMLP,
    BaseMLP,
    BetaMLP,
    create_alpha_mlp,
    create_beta_mlp,
)
from gravitorch.models.networks.mnist import PyTorchMnistNet



r"""This module contains some tools to analyze the parameters of a
``torch.nn.Module``."""

__all__ = ["compute_parameter_stats", "show_parameter_stats"]

import logging

from tabulate import tabulate
from torch.nn import Module

logger = logging.getLogger(__name__)


def compute_parameter_stats(module: Module) -> list[list]:
    r"""Computes the parameter statistics of a ``torch.nn.Module``.

    Args:
        module (``torch.nn.Module``): Specifies the module to analyze.

    Returns:
        list: The list of statistics per parameters.
    """
    stats = [["parameter", "mean", "median", "std", "min", "max", "learnable"]]
    for key, weight in module.named_parameters():
        weight = weight.flatten()
        if weight.numel() > 0:
            stats.append(
                [
                    key,
                    weight.mean().item(),
                    weight.median().item(),
                    weight.std(dim=0).item(),
                    weight.min().item(),
                    weight.max().item(),
                    weight.requires_grad,
                ]
            )
    return stats


def show_parameter_stats(module: Module, tablefmt: str = "rst") -> None:
    r"""Shows some statistics about the model parameters.

    Args:
        module (``torch.nn.Module``): Specifies the module to analyze.
        tablefmt (str, optional): Specifies the table format.
            Default: ``'rst'``
    """
    stats = compute_parameter_stats(module)
    logger.info(
        "Parameters statistics\n"
        f'{tabulate(stats, headers="firstrow", tablefmt=tablefmt, floatfmt=".6f")}\n'
    )



r"""This package contains the optimizer base class and some implemented
optimizers."""

__all__ = ["NoOpOptimizer", "setup_optimizer"]

from gravitorch.optimizers.factory import setup_optimizer
from gravitorch.optimizers.noop import NoOpOptimizer



r"""This file implements a history manager."""

__all__ = ["HistoryManager"]

import copy
import logging
from typing import Any, Optional

from gravitorch.utils.format import str_add_indent, to_torch_mapping_str
from gravitorch.utils.history.base import BaseHistory
from gravitorch.utils.history.generic import GenericHistory
from gravitorch.utils.history.utils import get_best_values

logger = logging.getLogger(__name__)


class HistoryManager:
    r"""Implements a manager to easily manage a group of histories.

    This class proposes an approach to manage a group of histories, but
    it is possible to use other approaches. If this class does not fit
    your needs, feel free to use another approach.
    """

    def __init__(self):
        self._histories = {}

    def __len__(self) -> int:
        r"""int: The number of histories in the manager."""
        return len(self._histories)

    def __repr__(self) -> str:
        if self._histories:
            return (
                f"{self.__class__.__qualname__}(\n"
                f"  {str_add_indent(to_torch_mapping_str(self._histories))}\n)"
            )
        return f"{self.__class__.__qualname__}()"

    def add_history(self, history: BaseHistory, key: Optional[str] = None) -> None:
        r"""Adds a history to the manager.

        Args:
            history (``BaseHistory``): Specifies the history
                to add to the manager.
            key (str or ``None``, optional): Specifies the key to
                store the history. If ``None``, the name of the
                history is used. Default: ``None``

        Example:

        .. code-block:: python

            >>> from gravitorch.utils.history import HistoryManager, MinScalarHistory
            >>> manager = HistoryManager()
            >>> manager.add_history(MinScalarHistory("loss"))
            >>> manager
            HistoryManager(
              (loss) MinScalarHistory(name=loss, max_size=10, history=())
            )
            >>> manager.add_history(MinScalarHistory("loss"), 'my key')
            HistoryManager(
              (loss) MinScalarHistory(name=loss, max_size=10, history=())
              (my key) MinScalarHistory(name=loss, max_size=10, history=())
            )
        """
        if key is None:
            key = history.name
        if key in self._histories:
            logger.warning(
                f"The {key} history already exists and will be replace by the new history"
            )
        self._histories[key] = history

    def get_best_values(self, prefix: str = "", suffix: str = "") -> dict[str, Any]:
        """Gets the best value of each metric.

        This method ignores the metrics with empty history and the
        non-comparable history.

        Args:
            prefix (str): Specifies the prefix used to create the dict
                of best values. The goal of this prefix is to generate
                a name which is different from the metric name to
                avoid confusion. By default, the returned dict uses the
                same name as the metric. Default: ``''``
            suffix (str): Specifies the suffix used to create the dict
                of best values. The goal of this suffix is to generate
                a name which is different from the metric name to
                avoid confusion. By default, the returned dict uses the
                same name as the metric. Default: ``''``

        Returns:
            dict: The dict with the best value of each metric.

        Example:

        .. code-block:: python

            >>> from gravitorch.utils.history import HistoryManager
            >>> manager = HistoryManager()
            >>> manager.get_best_values()
            {'accuracy': 42}
            >>> manager.get_best_values(prefix='best/')
            {'best/accuracy': 42}
            >>> manager.get_best_values(suffix='/best')
            {'accuracy/best': 42}
        """
        return get_best_values(self._histories, prefix=prefix, suffix=suffix)

    def get_history(self, key: str) -> BaseHistory:
        r"""Gets the history associated to a key.

        Args:
            key (str): Specifies the key of the history to retrieve.

        Returns
            ``BaseHistory``: The history if it exists,
                otherwise it returns an empty history. The created
                empty history is a ``GenericHistory``.

        Example:

        .. code-block:: python

            >>> from gravitorch.utils.history import HistoryManager, MinScalarHistory
            >>> manager = HistoryManager()
            >>> from gravitorch.utils.history import
            >>> manager.add_history(MinScalarHistory("loss"))
            >>> manager.get_history('loss')
            MinScalarHistory(name='loss', ...)
            >>> manager.get_history('new_history')
            GenericHistory(name='new_history', ...)
        """
        if not self.has_history(key):
            self._histories[key] = GenericHistory(name=key)
        return self._histories[key]

    def get_histories(self) -> dict[str, BaseHistory]:
        r"""Gets all the histories.

        Returns
            ``dict``: The histories with their keys.

        Example:

        .. code-block:: python

            >>> from gravitorch.utils.history import HistoryManager, MinScalarHistory
            >>> manager = HistoryManager()
            >>> manager.add_history(MinScalarHistory("loss"))
            >>> manager.get_histories()
            {'loss': MinScalarHistory(name='loss', ...)}
        """
        return copy.copy(self._histories)

    def has_history(self, key: str) -> bool:
        r"""Indicates if the engine has a history for the given key.

        Args:
            key (str): Specifies the key of the history.

        Returns
            bool: ``True`` if the history exists, ``False`` otherwise

        Example:

        .. code-block:: python

            >>> from gravitorch.utils.history import HistoryManager, MinScalarHistory
            >>> manager = HistoryManager()
            >>> manager.add_history(MinScalarHistory("loss"))
            >>> manager.has_history('loss')
            True
            >>> manager.has_history('missing')
            False
        """
        return key in self._histories

    def load_state_dict(self, state_dict: dict) -> None:
        r"""Loads the state values from a dict.

        Args:
            state_dict (dict): a dict with parameters

        Example:

        .. code-block:: python

            >>> from gravitorch.utils.history import HistoryManager
            >>> manager = HistoryManager()
            >>> my_state_dict = {...}  # the state dict depends on the histories
            >>> manager.load_state_dict(my_state_dict)
        """
        for key, state in state_dict.items():
            if self.has_history(key):
                self._histories[key].load_state_dict(state["state"])
            else:
                self._histories[key] = BaseHistory.from_dict(state)

    def state_dict(self) -> dict:
        r"""Returns a dictionary containing state values of all the histories.

        Returns:
            dict: the state values in a dict.

        Example:

        .. code-block:: python

            >>> from gravitorch.utils.history import HistoryManager
            >>> manager = HistoryManager()
            >>> manager.state_dict()
            {}
        """
        return {key: hist.to_dict() for key, hist in self._histories.items()}



r"""This package contains the implementation of some evaluation loops."""

__all__ = [
    "BaseHistory",
    "ComparableHistory",
    "EmptyHistoryError",
    "GenericHistory",
    "HistoryManager",
    "MaxScalarHistory",
    "MinScalarHistory",
    "NotAComparableHistoryError",
    "get_best_values",
    "get_last_values",
]

from gravitorch.utils.history.base import (
    BaseHistory,
    EmptyHistoryError,
    NotAComparableHistoryError,
)
from gravitorch.utils.history.comparable import (
    ComparableHistory,
    MaxScalarHistory,
    MinScalarHistory,
)
from gravitorch.utils.history.generic import GenericHistory
from gravitorch.utils.history.manager import HistoryManager
from gravitorch.utils.history.utils import get_best_values, get_last_values



r"""This module defines the base class for the data loader collators."""

__all__ = ["BaseCollator"]

from abc import ABC, abstractmethod
from collections.abc import Callable
from typing import Generic, TypeVar

from objectory import AbstractFactory

R = TypeVar("R")
T = TypeVar("T")


class BaseCollator(Generic[T, R], Callable[[list[T]], R], ABC, metaclass=AbstractFactory):
    r"""Defines the base class to create a batch of examples."""

    @abstractmethod
    def __call__(self, data: list[T]) -> R:
        r"""Creates a batch given a list of examples.

        Args:
            data (list): Specifies a list of examples.

        Returns:
             A batch of examples.
        """



__all__ = ["EpochOptimizerMonitor", "IterationOptimizerMonitor"]

import logging

from gravitorch.engines.base import BaseEngine
from gravitorch.engines.events import EngineEvents
from gravitorch.handlers.base import BaseHandler
from gravitorch.handlers.utils import add_unique_event_handler
from gravitorch.optimizers.utils import (
    log_optimizer_parameters_per_group,
    show_optimizer_parameters_per_group,
)
from gravitorch.utils.events import (
    ConditionalEventHandler,
    EpochPeriodicCondition,
    IterationPeriodicCondition,
)
from gravitorch.utils.exp_trackers import EpochStep, IterationStep

logger = logging.getLogger(__name__)


class EpochOptimizerMonitor(BaseHandler):
    r"""Implements a handler to monitor the optimizer every ``freq`` epochs.

    Args:
        event (str, optional): Specifies the epoch-based event when
            the optimizer information should be capture.
            Default: ``'train_epoch_started'``
        freq (int, optional): Specifies the epoch frequency used to
            monitor the optimizer. Default: ``1``
        tablefmt (str, optional): Specifies the table format to show
            the optimizer information. You can find the valid formats
            at https://pypi.org/project/tabulate/.
            Default: ``'fancy_grid'``
        prefix (str, optional): Specifies the prefix which is used to
            log metrics. Default: ``"train/"``
    """

    def __init__(
        self,
        event: str = EngineEvents.TRAIN_EPOCH_STARTED,
        freq: int = 1,
        tablefmt: str = "fancy_grid",
        prefix: str = "train/",
    ):
        self._event = str(event)
        if freq < 1:
            raise ValueError(f"freq has to be greater than 0 (received: {freq:,})")
        self._freq = int(freq)
        self._tablefmt = str(tablefmt)
        self._prefix = str(prefix)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(event={self._event}, freq={self._freq}, "
            f"tablefmt={self._tablefmt}, prefix={self._prefix})"
        )

    def attach(self, engine: BaseEngine) -> None:
        add_unique_event_handler(
            engine=engine,
            event=self._event,
            event_handler=ConditionalEventHandler(
                self.monitor,
                condition=EpochPeriodicCondition(engine=engine, freq=self._freq),
                handler_kwargs={"engine": engine},
            ),
        )

    def monitor(self, engine: BaseEngine) -> None:
        r"""Monitors the current optimizer state.

        Args:
            engine (``BaseEngine``): Specifies the engine.
        """
        if engine.optimizer:
            show_optimizer_parameters_per_group(optimizer=engine.optimizer, tablefmt=self._tablefmt)
            log_optimizer_parameters_per_group(
                optimizer=engine.optimizer,
                engine=engine,
                step=EpochStep(engine.epoch),
                prefix=self._prefix,
            )
        else:
            logger.info(
                "It is not possible to monitor the optimizer parameters because there is no "
                "optimizer"
            )


class IterationOptimizerMonitor(BaseHandler):
    r"""Implements a handler to monitor the optimizer every ``freq`` iterations.

    Args:
        event (str, optional): Specifies the iteration-based event
            when the optimizer information should be capture.
            Default: ``'train_iteration_started'``
        freq (int, optional): Specifies the iteration frequency used
            to monitor the optimizer. Default: ``10``
        tablefmt (str, optional): Specifies the table format to show
            the optimizer information. You can find the valid formats
            at https://pypi.org/project/tabulate/.
            Default: ``'fancy_grid'``
        prefix (str, optional): Specifies the prefix which is used to
            log metrics. Default: ``"train/"``
    """

    def __init__(
        self,
        event: str = EngineEvents.TRAIN_ITERATION_STARTED,
        freq: int = 10,
        tablefmt: str = "fancy_grid",
        prefix: str = "train/",
    ):
        self._event = str(event)
        if freq < 1:
            raise ValueError(f"freq has to be greater than 0 (received: {freq:,})")
        self._freq = int(freq)
        self._tablefmt = str(tablefmt)
        self._prefix = str(prefix)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(event={self._event}, freq={self._freq}, "
            f"tablefmt={self._tablefmt}, prefix={self._prefix})"
        )

    def attach(self, engine: BaseEngine) -> None:
        add_unique_event_handler(
            engine=engine,
            event=self._event,
            event_handler=ConditionalEventHandler(
                self.monitor,
                condition=IterationPeriodicCondition(engine=engine, freq=self._freq),
                handler_kwargs={"engine": engine},
            ),
        )

    def monitor(self, engine: BaseEngine) -> None:
        r"""Monitors the current optimizer state.

        Args:
            engine (``BaseEngine``): Specifies the engine.
        """
        if engine.optimizer:
            show_optimizer_parameters_per_group(optimizer=engine.optimizer, tablefmt=self._tablefmt)
            log_optimizer_parameters_per_group(
                optimizer=engine.optimizer,
                engine=engine,
                step=IterationStep(engine.iteration),
                prefix=self._prefix,
            )
        else:
            logger.info(
                "It is not possible to monitor the optimizer parameters because there is no "
                "optimizer"
            )



r"""This module implements an experiment tracker using TensorBoard."""

__all__ = ["TensorBoardExpTracker"]

import logging
import shutil
import tempfile
from pathlib import Path
from typing import Any, Optional, Union
from unittest.mock import Mock

import torch

from gravitorch import constants as ct
from gravitorch.utils.artifacts import BaseArtifact
from gravitorch.utils.exp_trackers.base import (
    BaseBasicExpTracker,
    NotActivatedExpTrackerError,
)
from gravitorch.utils.exp_trackers.steps import Step
from gravitorch.utils.format import to_flat_dict
from gravitorch.utils.integrations import (
    check_tensorboard,
    is_matplotlib_available,
    is_pillow_available,
    is_tensorboard_available,
)
from gravitorch.utils.io import load_json, save_json
from gravitorch.utils.path import sanitize_path

if is_matplotlib_available():
    from matplotlib.pyplot import Figure
else:
    Figure = "matplotlib.pyplot.Figure"  # pragma: no cover


if is_pillow_available():
    from PIL.Image import Image
else:
    Image = "PIL.Image.Image"  # pragma: no cover


if is_tensorboard_available():
    from torch.utils.tensorboard import SummaryWriter
    from torch.utils.tensorboard.summary import hparams
else:
    SummaryWriter, hparams = Mock, None  # pragma: no cover

logger = logging.getLogger(__name__)

BEST_METRIC_FILENAME = "best_metric.json"
MODEL_FILENAME = "model.txt"
TENSORBOARD_FOLDER_NAME = "tb"


class TensorBoardExpTracker(BaseBasicExpTracker):
    r"""Implements an experiment tracker using TensorBoard.

    This module is under development and only supports some features.

    Args:
        experiment_path (``pathlib.Path`` or str or ``None``,
            optional): Specifies the path where to write the
            experiment logs. If ``None``, a temporary directory is
            used. The temporary directory will be removed at the end
            of the experiment. Default: ``None``
        remove_after_run (bool, optional): If ``True``, the experiment
            directory is deleted at the end of the run, otherwise the
            experiment directory is kept. This option is ignored if a
            temporary directory is used. Default: ``False``
    """

    def __init__(
        self,
        experiment_path: Union[Path, str, None] = None,
        remove_after_run: bool = False,
    ):
        check_tensorboard()
        self._experiment_path = sanitize_path(experiment_path) if experiment_path else None
        # Flag to indicate if the tracker is activated or not
        self._is_activated = False
        # Flag to indicate if the experiment directory should be removed at the end of the run.
        self._remove_after_run = bool(remove_after_run)

        # The following directories or filenames are defined in the start method
        self._checkpoint_path: Optional[Path] = None
        self._artifact_path: Optional[Path] = None
        self._best_metric_path: Optional[Path] = None
        # The writer is initialized in the start method
        self._writer = None

        self._best_metrics = {}
        self._hparams = {}

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(experiment_path={self._experiment_path}, "
            f"is_activated={self._is_activated})"
        )

    @property
    def artifact_path(self) -> Path:
        if not self._is_activated:
            raise NotActivatedExpTrackerError(
                "It is not possible to get `artifact_path` because the experiment tracker "
                "is not activated"
            )
        return self._artifact_path

    @property
    def checkpoint_path(self) -> Path:
        if not self._is_activated:
            raise NotActivatedExpTrackerError(
                "It is not possible to get `checkpoint_path` because the experiment tracker "
                "is not activated"
            )
        return self._checkpoint_path

    @property
    def experiment_id(self) -> str:
        if not self._is_activated:
            raise NotActivatedExpTrackerError(
                "It is not possible to get `experiment_id` because the experiment tracker "
                "is not activated"
            )
        return "fakeid0123"

    def _start(self) -> None:
        self._is_activated = True
        if self._experiment_path is None:
            self._experiment_path = sanitize_path(tempfile.mkdtemp())
            self._remove_after_run = True

        logger.info(
            f"Create a TensorBoard experiment tracker and store results at {self._experiment_path}"
        )
        self._best_metric_path = self._experiment_path.joinpath(BEST_METRIC_FILENAME)
        self._writer = MLTorchSummaryWriter(
            log_dir=self._experiment_path.joinpath(TENSORBOARD_FOLDER_NAME).as_posix()
        )
        self._artifact_path = self._experiment_path.joinpath(ct.ARTIFACT_FOLDER_NAME)
        self._checkpoint_path = self._experiment_path.joinpath(ct.CHECKPOINT_FOLDER_NAME)

        self._load_best_metrics()

    def _flush(self, upload_checkpoints: bool = True) -> None:
        self._save_best_metrics()
        self._writer.add_hparams(_sanitize_dict(self._hparams), self._best_metrics)
        self._writer.flush()
        if upload_checkpoints:
            self.upload_checkpoints()

    def _end(self) -> None:
        logger.info("Stopping the TensorBoard experiment tracker...")
        self.flush()
        logger.info(f"::::: TensorBoard logs are available at {self._experiment_path} :::::")
        self._writer.close()
        self._writer = None
        self._best_metrics.clear()
        self._hparams.clear()

        if self._remove_after_run:
            # Remove the temporary directory
            shutil.rmtree(self._experiment_path)
            self._experiment_path = None

        self._artifact_path = None
        self._checkpoint_path = None
        self._is_activated = False
        self._best_metric_path = None

    def is_activated(self) -> bool:
        return self._is_activated

    def _is_resumed(self) -> bool:
        return False  # TODO

    def _add_tag(self, name: str, value: Any) -> None:
        pass  # Do nothing because there is no equivalent feature in TensorBoard

    def _add_tags(self, tags: dict[str, Any]) -> None:
        pass  # Do nothing because there is no equivalent feature in TensorBoard

    def _create_artifact(self, artifact: BaseArtifact) -> None:
        artifact.create(self.artifact_path)

    def _log_best_metric(self, key: str, value: Union[int, float]) -> None:
        self._best_metrics[f"{key}.best"] = _sanitize_value(value)

    def _log_best_metrics(self, metrics: dict[str, Union[int, float]]) -> None:
        for key, value in metrics.items():
            self._log_best_metric(key, value)

    def _log_figure(self, key: str, figure: Figure, step: Optional[Step] = None) -> None:
        self._writer.add_figure(key, figure, step.step if step else step)

    def _log_figures(self, figures: dict[str, Figure], step: Optional[Step] = None) -> None:
        step = step.step if step else step
        for key, figure in figures.items():
            self._writer.add_figure(key, figure, step)

    def _log_hyper_parameter(self, key: str, value: Any) -> None:
        self._hparams[key] = value

    def _log_hyper_parameters(self, params: dict[str, Any]) -> None:
        self._hparams.update(params)

    def _log_image(self, key: str, image: Image, step: Optional[Step] = None) -> None:
        pass  # Do nothing because there is no equivalent feature in TensorBoard

    def _log_images(self, images: dict[str, Image], step: Optional[Step] = None) -> None:
        pass  # Do nothing because there is no equivalent feature in TensorBoard

    def _log_metric(self, key: str, value: Union[int, float], step: Optional[Step] = None) -> None:
        self._writer.add_scalar(key, _sanitize_value(value), step.step if step else step)

    def _log_metrics(
        self, metrics: dict[str, Union[int, float]], step: Optional[Step] = None
    ) -> None:
        step = step.step if step else step
        for key, value in _sanitize_dict(metrics).items():
            self._writer.add_scalar(key, value, step)

    def _upload_checkpoints(self) -> None:
        pass  # Do nothing because it is not possible to upload checkpoints to TensorBoard

    def _load_best_metrics(self) -> None:
        r"""Loads the best metrics from a JSON file."""
        if self._best_metric_path.is_file():
            self._best_metrics = load_json(self._best_metric_path)

    def _save_best_metrics(self) -> None:
        r"""Saves the best metrics in a JSON file."""
        save_json(self._best_metrics, self._best_metric_path)


class MLTorchSummaryWriter(SummaryWriter):
    r"""Implements a variant of the ``SummaryWriter``.

    The goal is to refine ``add_hparams`` to be able to call it several
    times per experiment. Otherwise, a new experiment is created every
    time thath/issues/32651#issuecomment-643791116
    orch/issues/32651#issuecomment-648340103
    https://github.com/pytorch/pyt
    """

    def add_hparams(self, hparam_dict, metric_dict):
        r"""Add a set of hyperparameters to be compared in TensorBoard.

        Args:
            hparam_dict (dict): Each key-value pair in the dictionary is the
              name of the hyper parameter and it's corresponding value.
              The type of the value can be one of `bool`, `string`, `float`,
              `int`, or `None`.
            metric_dict (dict): Each key-value pair in the dictionary is the
              name of the metric and it's corresponding value. Note that the key used
              here should be unique in the tensorboard record. Otherwise the value
              you added by ``add_scalar`` will be displayed in hparam plugin. In most
              cases, this is unwanted.

        Examples::

            from torch.utils.tensorboard import SummaryWriter
            with SummaryWriter() as w:
                for i in range(5):
                    w.add_hparams({'lr': 0.1*i, 'bsize': i},
                                  {'hparam/accuracy': 10*i, 'hparam/loss': 10*i})
        """
        torch._C._log_api_usage_once("tensorboard.logging.add_hparams")
        if type(hparam_dict) is not dict or type(metric_dict) is not dict:
            raise TypeError("hparam_dict and metric_dict should be dictionary.")
        exp, ssi, sei = hparams(hparam_dict, metric_dict)

        self.file_writer.add_summary(exp)
        self.file_writer.add_summary(ssi)
        self.file_writer.add_summary(sei)
        for k, v in metric_dict.items():
            self.add_scalar(k, v)


def _sanitize_dict(hparams: dict[str, Any]) -> dict[str, Any]:
    """Computes a sanitized version of the hyper-parameters.

    The values that are not compatible with TensorBoard are converted
    to a string.

    Args:
        hparams (dict): Specifies the hyper-parameters to sanitize.

    Returns:
        dict: The sanitized hyper-parameters.
    """
    return {key: _sanitize_value(value) for key, value in to_flat_dict(hparams).items()}


def _sanitize_value(value: Any) -> Union[bool, int, float, str]:
    """Computes a sanitized a value.

    The values that are not compatible with TensorBoard are converted
    to a string.

    Args:
        value (dict): Specifies the value to sanitize.

    Returns:
        dict: The sanitized value.
    """
    if isinstance(value, (bool, int, float, str)):
        return value
    elif torch.is_tensor(value) and value.numel() == 1:
        return value
    return str(value)



__all__ = ["NoOpLoopObserver"]

from typing import Any

from gravitorch.engines.base import BaseEngine
from gravitorch.utils.loop_observers.base import BaseLoopObserver


class NoOpLoopObserver(BaseLoopObserver):
    r"""Implements a no-operation loop observer."""

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def start(self, engine: BaseEngine) -> None:
        r"""It is a no-operation method."""

    def end(self, engine: BaseEngine) -> None:
        r"""It is a no-operation method."""

    def update(self, engine: BaseEngine, model_input: Any, model_output: Any) -> None:
        r"""It is a no-operation method."""



__all__ = ["VanillaCoreCreator"]

from typing import Optional, Union

from torch import nn
from torch.optim import Optimizer

from gravitorch import constants as ct
from gravitorch.creators.core.base import BaseCoreCreator
from gravitorch.datasources.base import BaseDataSource
from gravitorch.datasources.utils import setup_data_source
from gravitorch.engines.base import BaseEngine
from gravitorch.lr_schedulers.base import LRSchedulerType, setup_lr_scheduler
from gravitorch.models.utils import setup_model
from gravitorch.optimizers.factory import setup_optimizer
from gravitorch.utils.format import str_add_indent


class VanillaCoreCreator(BaseCoreCreator):
    r"""Implements a simple core engine moules creator.

    This creator does not always "create" the core modules because
    they can already exist. The user is responsible to attach the
    core modules to the engine. This creator only adds the given
    modules to the engine state.

    Args:
        data_source (``BaseDataSource`` or dict): Specifies the data
            source or its configuration.
        model (``BaseModelCreator`` or dict): Specifies the model
            or its configuration.
        optimizer (``BaseOptimizerCreator`` or dict or ``None`):
            Specifies the optimizer or its configuration.
            Default: ``None``
        lr_scheduler (``BaseLRSchedulerCreator`` or dict or ``None`):
            Specifies the LR scheduler or its configuration.
            Default: ``None``
    """

    def __init__(
        self,
        data_source: Union[BaseDataSource, dict],
        model: Union[nn.Module, dict],
        optimizer: Union[Optimizer, dict, None] = None,
        lr_scheduler: Union[LRSchedulerType, dict, None] = None,
    ):
        self._data_source = setup_data_source(data_source)
        self._model = setup_model(model)
        self._optimizer = setup_optimizer(model=self._model, optimizer=optimizer)
        self._lr_scheduler = setup_lr_scheduler(
            optimizer=self._optimizer, lr_scheduler=lr_scheduler
        )

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  data_source={str_add_indent(self._data_source)},\n"
            f"  model={str_add_indent(self._model)},\n"
            f"  optimizer={str_add_indent(self._optimizer)},\n"
            f"  lr_scheduler={str_add_indent(self._lr_scheduler)},\n"
            ")"
        )

    def create(
        self, engine: BaseEngine
    ) -> tuple[BaseDataSource, nn.Module, Optional[Optimizer], Optional[LRSchedulerType]]:
        engine.add_module(ct.DATA_SOURCE, self._data_source)
        engine.add_module(ct.MODEL, self._model)
        if self._optimizer:
            engine.add_module(ct.OPTIMIZER, self._optimizer)
        if self._lr_scheduler:
            engine.add_module(ct.LR_SCHEDULER, self._lr_scheduler)
        return self._data_source, self._model, self._optimizer, self._lr_scheduler



r"""This package implements some fusion layers."""

__all__ = [
    "AverageFusion",
    "ConcatFusion",
    "FusionFFN",
    "FusionNorm",
    "MultiplicationFusion",
    "SumFusion",
]

from gravitorch.nn.fusion.concat import ConcatFusion
from gravitorch.nn.fusion.ffn import FusionFFN
from gravitorch.nn.fusion.multiplication import MultiplicationFusion
from gravitorch.nn.fusion.normalization import FusionNorm
from gravitorch.nn.fusion.sum import AverageFusion, SumFusion



r"""This module implements a simple evaluation loop."""

__all__ = ["VanillaEvaluationLoop"]

import logging
import sys
from collections.abc import Iterable
from typing import Any, Union

import torch
from torch.nn import Module
from tqdm import tqdm

from gravitorch.distributed import comm as dist
from gravitorch.engines.base import BaseEngine
from gravitorch.engines.events import EngineEvents
from gravitorch.utils.device_placement import (
    AutoDevicePlacement,
    BaseDevicePlacement,
    setup_device_placement,
)
from gravitorch.utils.evaluation_loops.base import BaseBasicEvaluationLoop
from gravitorch.utils.evaluation_loops.conditions import BaseEvalCondition
from gravitorch.utils.loop_observers import BaseLoopObserver
from gravitorch.utils.profilers import BaseProfiler

logger = logging.getLogger(__name__)


class VanillaEvaluationLoop(BaseBasicEvaluationLoop):
    r"""Implements a simple evaluation loop to evaluate a model on a given
    dataset.

    Args:
        grad_enabled (bool, optional): Specifies if the gradient is
            computed or not in the evaluation loop. By default, the
            gradient is not computed to reduce the memory footprint.
            Default: ``False``
        batch_device_placement (``BaseDevicePlacement`` or dict or
            ``None``, optional): Specifies the batch device placement
            module. This module moves the batch on a target device.
            The target device should be compatible with the model.
            If ``None``, an ``AutoDevicePlacement`` object is
            instantiated. Default: ``None``
        tag (str, optional): Specifies the tag which is used to log
            metrics. Default: ``"eval"``
        condition (``BaseEvalCondition`` or dict or None): Specifies
            the condition to evaluate the loop or its configuration.
            If ``None``, the ``EveryEpochEvalCondition(every=1)`` is
            used.  Default ``None``
        observer (``BaseLoopObserver`` or dict or None, optional):
            Specifies the loop observer or its configuration.
            If ``None``, the ``NoOpLoopObserver`` is instantiated.
            Default: ``None``
        profiler (``BaseProfiler`` or dict or None, optional):
            Specifies the profiler or its configuration. If ``None``,
            the ``NoOpProfiler`` is instantiated. Default: ``None``
    """

    def __init__(
        self,
        grad_enabled: bool = False,
        batch_device_placement: Union[BaseDevicePlacement, dict, None] = None,
        tag: str = "eval",
        condition: Union[BaseEvalCondition, dict, None] = None,
        observer: Union[BaseLoopObserver, dict, None] = None,
        profiler: Union[BaseProfiler, dict, None] = None,
    ):
        super().__init__(tag=tag, condition=condition, observer=observer, profiler=profiler)
        self._grad_enabled = bool(grad_enabled)
        self._batch_device_placement = setup_device_placement(
            batch_device_placement or AutoDevicePlacement()
        )

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  tag={self._tag},\n"
            f"  batch_device_placement={self._batch_device_placement},\n"
            f"  grad_enabled={self._grad_enabled},\n"
            f"  condition={self._condition},\n"
            f"  observer={self._observer},\n"
            f"  profiler={self._profiler},\n"
            ")"
        )

    def _eval_one_batch(self, engine: BaseEngine, model: Module, batch: Any) -> dict:
        engine.fire_event(EngineEvents.EVAL_ITERATION_STARTED)
        with torch.set_grad_enabled(self._grad_enabled):
            output = model(self._batch_device_placement.send(batch))
        engine.fire_event(EngineEvents.EVAL_ITERATION_COMPLETED)
        return output

    def _prepare_model_data_loader(self, engine: BaseEngine) -> tuple[Module, Iterable]:
        logger.info("Preparing the model and data loader...")
        data_loader = engine.data_source.get_data_loader(loader_id=self._tag, engine=engine)
        prefix = f"({dist.get_rank()}/{dist.get_world_size()}) " if dist.is_distributed() else ""
        data_loader = tqdm(
            data_loader,
            desc=f"{prefix}Evaluation [{engine.epoch}]",
            position=dist.get_rank(),
            file=sys.stdout,
        )
        logger.info("Evaluation data loader has been created")
        return engine.model, data_loader



r"""This package contains the implementation of some evaluation loops."""

__all__ = [
    "AMPEvaluationLoop",
    "AccelerateEvaluationLoop",
    "BaseBasicEvaluationLoop",
    "BaseEvaluationLoop",
    "NoOpEvaluationLoop",
    "VanillaEvaluationLoop",
    "setup_evaluation_loop",
]

from gravitorch.utils.evaluation_loops.accelerate import AccelerateEvaluationLoop
from gravitorch.utils.evaluation_loops.amp import AMPEvaluationLoop
from gravitorch.utils.evaluation_loops.base import (
    BaseBasicEvaluationLoop,
    BaseEvaluationLoop,
)
from gravitorch.utils.evaluation_loops.noop import NoOpEvaluationLoop
from gravitorch.utils.evaluation_loops.utils import setup_evaluation_loop
from gravitorch.utils.evaluation_loops.vanilla import VanillaEvaluationLoop



__all__ = ["create_sequential_iter_datapipe", "setup_iter_datapipe"]

from collections.abc import Sequence
from typing import Union

from objectory import OBJECT_TARGET, factory
from torch.utils.data import IterDataPipe


def create_sequential_iter_datapipe(configs: Sequence[dict]) -> IterDataPipe:
    r"""Creates a sequence of ``IterDataPipe``s from their configuration.

    Args:
        configs: Specifies the config of each ``IterDataPipe`` in the
            sequence. The configs sequence follows the order of the
            ``IterDataPipe``s. The first config is used to create the
            first ``IterDataPipe`` (a.k.a. source), and the last
            config is used to create the last ``IterDataPipe``
            (a.k.a. sink). This function assumes all the DataPipes
            have a single source DataPipe as their first argument,
            excepts for the first DataPipe.

    Returns:
        ``IterDataPipe``: The last ``IterDataPipe`` of the sequence
            (a.k.a. sink).

    Raises:
        ValueError if the ``IterDataPipe`` configuration sequence is
            empty.

    Example usage:

    .. code-block:: python

        >>> from gravitorch.data.datapipes.iter.factory import create_sequential_iter_datapipe
        >>> datapipe = create_sequential_iter_datapipe(
        ...     [
        ...         {
        ...             '_target_': "gravitorch.data.datapipes.iter.SourceWrapper",
        ...             "data": [1, 2, 3, 4],
        ...         },
        ...     ],
        ... )
        >>> tuple(datapipe)
        (1, 2, 3, 4)
        >>> datapipe = create_sequential_iter_datapipe(
        ...     [
        ...         {
        ...             '_target_': "gravitorch.data.datapipes.iter.SourceWrapper",
        ...             "data": [1, 2, 3, 4],
        ...         },
        ...         {'_target_': "torch.utils.data.datapipes.iter.Batcher", "batch_size": 2},
        ...     ]
        ... )
        >>> tuple(datapipe)
        ([1, 2], [3, 4])
    """
    if not configs:
        raise ValueError("It is not possible to create a DataPipe because the configs are empty")
    datapipe = factory(**configs[0])
    for config in configs[1:]:
        config = config.copy()  # Make a copy because the dict is modified below.
        target = config.pop(OBJECT_TARGET)
        datapipe = factory(target, datapipe, **config)
    return datapipe


def setup_iter_datapipe(datapipe: Union[IterDataPipe, Sequence[dict]]) -> IterDataPipe:
    r"""Sets up an ``IterDataPipe``.

    Note: it is only possible to create a sequential DataPipe from
    its configuration. See ``create_sequential_iter_datapipe``
    documentation for more information.

    Args:
        datapipe: Specifies the DataPipe or its configuration.

    Returns:
        ``IterDataPipe``: The last ``IterDataPipe`` of the sequence
            (a.k.a. sink).

    Example usage:

    .. code-block:: python

        >>> from gravitorch.data.datapipes.iter import setup_iter_datapipe
        >>> datapipe = setup_iter_datapipe(
        ...     [
        ...         {
        ...             '_target_': "gravitorch.data.datapipes.iter.SourceWrapper",
        ...             "data": [1, 2, 3, 4],
        ...         },
        ...     ],
        ... )
        >>> tuple(datapipe)
        (1, 2, 3, 4)
        >>> datapipe = setup_iter_datapipe(
        ...     [
        ...         {
        ...             '_target_': "gravitorch.data.datapipes.iter.SourceWrapper",
        ...             "data": [1, 2, 3, 4],
        ...         },
        ...         {'_target_': "torch.utils.data.datapipes.iter.Batcher", "batch_size": 2},
        ...     ]
        ... )
        >>> tuple(datapipe)
        ([1, 2], [3, 4])
    """
    if isinstance(datapipe, IterDataPipe):
        return datapipe
    return create_sequential_iter_datapipe(datapipe)



r"""This package contains the implementation of the event system."""

__all__ = [
    "BaseEventHandler",
    "BaseEventHandlerWithArguments",
    "ConditionalEventHandler",
    "EpochPeriodicCondition",
    "EventManager",
    "IterationPeriodicCondition",
    "PeriodicCondition",
    "VanillaEventHandler",
]

from gravitorch.utils.events.conditions import (
    EpochPeriodicCondition,
    IterationPeriodicCondition,
    PeriodicCondition,
)
from gravitorch.utils.events.event_handlers import (
    BaseEventHandler,
    BaseEventHandlerWithArguments,
    ConditionalEventHandler,
    VanillaEventHandler,
)
from gravitorch.utils.events.manager import EventManager



__all__ = ["BaseDataSourceCreator", "VanillaDataSourceCreator", "setup_data_source_creator"]

from gravitorch.creators.datasource.base import (
    BaseDataSourceCreator,
    setup_data_source_creator,
)
from gravitorch.creators.datasource.vanilla import VanillaDataSourceCreator



__all__ = ["DictOfListConverterIterDataPipe", "ListOfDictConverterIterDataPipe"]

from collections.abc import Hashable, Iterator, Mapping, Sequence

from torch.utils.data import IterDataPipe

from gravitorch.utils.format import str_add_indent
from gravitorch.utils.mapping import convert_to_dict_of_lists, convert_to_list_of_dicts


class DictOfListConverterIterDataPipe(IterDataPipe[dict[Hashable, list]]):
    r"""Implements an ``IterDataPipe`` to convert a sequence of mappings to a
    dictionary of lists.

    Args:
        source_datapipe (``IterDataPipe``): Specifies an ``IterDataPipe``
            of sequences of mappings.
    """

    def __init__(self, source_datapipe: IterDataPipe[Sequence[Mapping]]):
        self._source_datapipe = source_datapipe

    def __iter__(self) -> Iterator[dict[Hashable, list]]:
        for data in self._source_datapipe:
            yield convert_to_dict_of_lists(data)

    def __len__(self) -> int:
        return len(self._source_datapipe)

    def __str__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  source_datapipe={str_add_indent(self._source_datapipe)},\n)"
        )


class ListOfDictConverterIterDataPipe(IterDataPipe[list[dict]]):
    r"""Implements an ``IterDataPipe`` to convert a mapping of sequences to a
    list of dictionaries.

    Args:
        source_datapipe (``IterDataPipe``): Specifies an
            ``IterDataPipe`` of mappings of sequences.
    """

    def __init__(self, source_datapipe: IterDataPipe[Mapping[Hashable, Sequence]]):
        self._source_datapipe = source_datapipe

    def __iter__(self) -> Iterator[list[dict]]:
        for data in self._source_datapipe:
            yield convert_to_list_of_dicts(data)

    def __len__(self) -> int:
        return len(self._source_datapipe)

    def __str__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  source_datapipe={str_add_indent(self._source_datapipe)},\n)"
        )



r"""This module implements some utility functions to use ``torch.Tensor``s."""

__all__ = [
    "UNKNOWN",
    "get_dtype",
    "get_shape",
    "recursive_apply",
    "recursive_contiguous",
    "recursive_detach",
    "recursive_transpose",
]

from collections.abc import Callable
from typing import Any, Optional, TypeVar

import torch

T = TypeVar("T")

UNKNOWN = "unknown"


def get_dtype(data: Any) -> Any:
    r"""Gets the tensor data type recursively in a data structure.

    The current implementation supports the following types:

        - ``collections.OrderedDict``
        - ``dict``
        - ``list``
        - ``torch.Tensor``
        - ``tuple``

    Args:
        data: Specifies the input data.

    Returns:
        An object with the same structure as the input but with the
            tensor data type.

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.tensor import get_dtype
        >>> get_dtype(torch.ones(1, 2, 3))
        torch.Size([1, 2, 3])
        >>> get_dtype({"key1": torch.ones(1, 2, 3), "key2": "abc"})
        {'key1': torch.Size([1, 2, 3]), 'key2': 'unknown'}
        >>> get_dtype([torch.ones(1, 2, 3), "abc"])
        [torch.Size([1, 2, 3]), 'unknown']
    """
    if torch.is_tensor(data):
        return data.dtype
    if isinstance(data, (list, tuple, set)):
        return type(data)(get_dtype(t) for t in data)
    if isinstance(data, dict):
        return type(data)({key: get_dtype(value) for key, value in data.items()})
    return UNKNOWN


def get_shape(data: Any) -> Any:
    r"""Gets the tensor shape recursively in a data structure.

    The current implementation supports the following types:

        - ``collections.OrderedDict``
        - ``dict``
        - ``list``
        - ``torch.Tensor``
        - ``tuple``

    Args:
        data: Specifies the input data.

    Returns:
        An object with the same structure as the input but with the
            tensor shapes.

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.tensor import get_shape
        >>> get_shape(torch.ones(1, 2, 3))
        torch.Size([1, 2, 3])
        >>> get_shape({"key1": torch.ones(1, 2, 3), "key2": "abc"})
        {'key1': torch.Size([1, 2, 3]), 'key2': 'unknown'}
        >>> get_shape([torch.ones(1, 2, 3), "abc"])
        [torch.Size([1, 2, 3]), 'unknown']
    """
    if torch.is_tensor(data):
        return data.shape
    if isinstance(data, (list, tuple, set)):
        return type(data)(get_shape(t) for t in data)
    if isinstance(data, dict):
        return type(data)({key: get_shape(value) for key, value in data.items()})
    return UNKNOWN


def recursive_apply(data: Any, tensor_fn: Callable, other_fn: Optional[Callable] = None) -> Any:
    r"""Recursively applies a function on all the ``torch.Tensor``s.

    The current implementation supports the following types:

        - ``collections.OrderedDict``
        - ``dict``
        - ``list``
        - ``set``
        - ``torch.Tensor``
        - ``tuple``

    Args:
        data: Specifies the data.
        tensor_fn (callable): Specifies the function to apply on
            ``torch.Tensor``s.
        other_fn (callable or ``None``, optional): Specifies the
            function to apply on non ``torch.Tensor`` values.
            ``None`` means the original value is returned.
            Default: ``None``

    Returns:
        The transformed data.

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.utils.tensor import recursive_apply
        >>> recursive_apply(torch.ones(2, 3), lambda tensor: tensor.dtype)
        torch.float32
        >>> recursive_apply(
        ...     {
        ...         'key1': torch.ones(2, 3),
        ...         'key2': 'abc',
        ...         'key3': [torch.ones(2, 3), torch.arange(6)]
        ...     },
        ...     lambda tensor: tensor.dtype,
        ... )
        {'key1': torch.float32, 'key2': 'abc', 'key3': [torch.float32, torch.int64]}
    """
    if torch.is_tensor(data):
        return tensor_fn(data)
    if isinstance(data, (list, tuple, set)):
        return type(data)(recursive_apply(t, tensor_fn, other_fn) for t in data)
    if isinstance(data, dict):
        return type(data)(
            {key: recursive_apply(value, tensor_fn, other_fn) for key, value in data.items()}
        )
    return other_fn(data) if other_fn else data


def recursive_contiguous(
    data: T, memory_format: torch.memory_format = torch.contiguous_format
) -> T:
    r"""Returns contiguous in memory tensors containing the same data as the
    input.

    The current implementation supports the following types:

        - ``collections.OrderedDict``
        - ``dict``
        - ``list``
        - ``set``
        - ``torch.Tensor``
        - ``tuple``

    Args:
        data: Specifies the data to transform tocontiguous in memory
            tensors. All the tensors will be transformed by using the
            ``contiguous`` method.
        memory_format (``torch.memory_format``, optional): Specifies
            the desired memory format.
            Default: ``torch.contiguous_format``

    Returns:
        Contiguous in memory tensors containing the same data as the
            input.

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.tensor import recursive_contiguous
        >>> x = torch.ones(2, 3).transpose(0, 1)
        >>> x.is_contiguous()
        False
        >>> recursive_contiguous(x).is_contiguous()
        True
        >>> out = recursive_contiguous({'key': x})
        >>> out['key'].is_contiguous()
        True
    """
    return recursive_apply(data, lambda tensor: tensor.contiguous(memory_format=memory_format))


def recursive_detach(data: T) -> T:
    r"""Detaches the ``torch.Tensor``s.

    The current implementation supports the following types:

        - ``collections.OrderedDict``
        - ``dict``
        - ``list``
        - ``set``
        - ``torch.Tensor``
        - ``tuple``

    Args:
        data: Specifies the data to detach. All the tensors will be
            detached by using the ``detach`` method.

    Returns:
        The detached data.

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.tensor import recursive_detach
        >>> recursive_detach(torch.ones(2, 3, requires_grad=True)).requires_grad
        False
        >>> out = recursive_detach({'key': torch.ones(2, 3, requires_grad=True)})
        >>> out['key'].requires_grad
        False
    """
    return recursive_apply(data, lambda tensor: tensor.detach())


def recursive_transpose(data: T, dim0: int, dim1: int) -> T:
    r"""Transposes all the ``torch.Tensor``s in the input.

    The current implementation supports the following types:

        - ``collections.OrderedDict``
        - ``dict``
        - ``list``
        - ``set``
        - ``torch.Tensor``
        - ``tuple``

    Note: all the tensors should be compatible with the transpose
    dimensions.

    Args:
        data: Specifies the data to transpose. All the
            ``torch.Tensor``s are transposed by using the
            ``transpose`` method.
        dim0 (int): Specifies the first dimension to be transposed.
        dim1 (int): Specifies the second dimension to be transposed.

    Returns:
        The input data where the ``torch.Tensor``s are transposed.

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.tensor import recursive_transpose
        >>> recursive_transpose(torch.ones(3, 2), 0, 1).shape
        torch.Size([2, 3])
        >>> out = recursive_transpose({'key': torch.ones(3, 2)})
        >>> out['key'].shape
        torch.Size([2, 3])
    """
    return recursive_apply(data, lambda tensor: tensor.transpose(dim0, dim1))



__all__ = ["ReLUn"]

from torch import Tensor
from torch.nn import Module


class ReLUn(Module):
    r"""Implements the ReLU-n module.

    ``ReLUn(x, n)=min(max(0,x),n)``

    Args:
        max_value (float, optional): Specifies the maximum value.
            Default: ``1.0``
    """

    def __init__(self, max_value: float = 1.0):
        super().__init__()
        self._max_value = float(max_value)

    def extra_repr(self) -> str:
        return f"max_value={self._max_value}"

    def forward(self, tensor: Tensor) -> Tensor:
        r"""Applies the element-wise ReLU-n function.

        Args:
            tensor (``torch.Tensor`` of shape ``(*)``): Specifies the
                input tensor.

        Returns:
            ``torch.Tensor`` with same shape as the input: The output
                tensor.
        """
        return tensor.clamp(min=0.0, max=self._max_value)



__all__ = ["VanillaEngineState"]

from typing import Any, Optional

from gravitorch.utils.engine_states.base import BaseEngineState
from gravitorch.utils.format import str_add_indent
from gravitorch.utils.history import BaseHistory, HistoryManager
from gravitorch.utils.module_manager import ModuleManager


class VanillaEngineState(BaseEngineState):
    r"""Defines the vanilla engine state.

    Args:
        epoch (int, optional): Specifies the number of epochs
            performed. Default: ``-1``.
        iteration (int, optional): Specifies the number of training
            iterations performed. Default: ``-1``.
        max_epochs (int, optional): Specifies the maximum number of
            epochs. Default: ``1``.
        random_seed (int, optional): Specifies the random seed.
            Default: ``9984043075503325450``.
    """

    def __init__(
        self,
        epoch: int = -1,
        iteration: int = -1,
        max_epochs: int = 1,
        random_seed: int = 9984043075503325450,
    ):
        self._epoch = epoch
        self._iteration = iteration
        self._max_epochs = max_epochs
        self._random_seed = random_seed

        self._histories = HistoryManager()
        self._modules = ModuleManager()

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  modules={str_add_indent(self._modules)},\n"
            f"  histories={str_add_indent(self._histories)},\n"
            f"  random_seed={self._random_seed},\n"
            f"  max_epochs={self._max_epochs},\n"
            f"  epoch={self._epoch},\n"
            f"  iteration={self._iteration},\n"
            ")"
        )

    @property
    def epoch(self) -> int:
        return self._epoch

    @property
    def iteration(self) -> int:
        return self._iteration

    @property
    def max_epochs(self) -> int:
        return self._max_epochs

    @property
    def random_seed(self) -> int:
        return self._random_seed

    def add_history(self, history: BaseHistory, key: Optional[str] = None) -> None:
        self._histories.add_history(history=history, key=key)

    def add_module(self, name: str, module: Any) -> None:
        self._modules.add_module(name=name, module=module)

    def get_best_values(self, prefix: str = "", suffix: str = "") -> dict[str, Any]:
        return self._histories.get_best_values(prefix, suffix)

    def get_history(self, key: str) -> BaseHistory:
        return self._histories.get_history(key)

    def get_histories(self) -> dict[str, BaseHistory]:
        return self._histories.get_histories()

    def get_module(self, name: str) -> Any:
        return self._modules.get_module(name)

    def has_history(self, key: str) -> bool:
        return self._histories.has_history(key)

    def has_module(self, name: str) -> bool:
        return self._modules.has_module(name)

    def increment_epoch(self, increment: int = 1) -> None:
        self._epoch += increment

    def increment_iteration(self, increment: int = 1) -> None:
        self._iteration += increment

    def load_state_dict(self, state_dict: dict) -> None:
        self._epoch = state_dict["epoch"]
        self._iteration = state_dict["iteration"]
        self._histories.load_state_dict(state_dict["histories"])
        self._modules.load_state_dict(state_dict["modules"])

    def remove_module(self, name: str) -> None:
        self._modules.remove_module(name)

    def state_dict(self) -> dict:
        return {
            "epoch": self._epoch,
            "iteration": self._iteration,
            "histories": self._histories.state_dict(),
            "modules": self._modules.state_dict(),
        }



__all__ = ["setup_device_placement"]

import logging
from typing import Union

from gravitorch.utils.device_placement.base import BaseDevicePlacement
from gravitorch.utils.device_placement.noop import NoOpDevicePlacement
from gravitorch.utils.format import str_target_object

logger = logging.getLogger(__name__)


def setup_device_placement(
    device_placement: Union[BaseDevicePlacement, dict, None]
) -> BaseDevicePlacement:
    r"""Sets up a device placement module.

    The device placement module is instantiated from its configuration
    by using the ``BaseDevicePlacement`` factory function.

    Args:
        device_placement (``BaseDevicePlacement`` or dict or ``None``):
            Specifies the device placement module or its configuration.
            If ``None``, the ``NoOpDevicePlacement`` is instantiated.

    Returns:
        ``BaseDevicePlacement``: The device placement module.
    """
    if device_placement is None:
        device_placement = NoOpDevicePlacement()
    if isinstance(device_placement, dict):
        logger.info(
            "Initializing a device placement module from its configuration... "
            f"{str_target_object(device_placement)}"
        )
        device_placement = BaseDevicePlacement.factory(**device_placement)
    return device_placement



r"""This module implements some utility functions for the evaluation loops."""

__all__ = ["setup_evaluation_loop"]

import logging
from typing import Union

from gravitorch.utils.evaluation_loops.base import BaseEvaluationLoop
from gravitorch.utils.evaluation_loops.vanilla import VanillaEvaluationLoop
from gravitorch.utils.format import str_target_object

logger = logging.getLogger(__name__)


def setup_evaluation_loop(
    evaluation_loop: Union[BaseEvaluationLoop, dict, None]
) -> BaseEvaluationLoop:
    r"""Sets up the evaluation loop.

    The evaluation loop is instantiated from its configuration by
    using the ``BaseEvaluationLoop`` factory function.

    Args:
        evaluation_loop (``BaseEvaluationLoop`` or dict or None):
            Specifies the evaluation loop or its configuration.
            If ``None``, the ``VanillaEvaluationLoop`` is instantiated.

    Returns:
        ``BaseEvaluationLoop``: The evaluation loop.
    """
    if evaluation_loop is None:
        evaluation_loop = VanillaEvaluationLoop()
    if isinstance(evaluation_loop, dict):
        logger.info(
            f"Initializing an evaluation loop from its configuration... {str_target_object(evaluation_loop)}"
        )
        evaluation_loop = BaseEvaluationLoop.factory(**evaluation_loop)
    return evaluation_loop



r"""This module implements a no-operation evaluation loop."""

__all__ = ["NoOpEvaluationLoop"]

import logging

from gravitorch.engines.base import BaseEngine
from gravitorch.utils.evaluation_loops.base import BaseEvaluationLoop

logger = logging.getLogger(__name__)


class NoOpEvaluationLoop(BaseEvaluationLoop):
    r"""Implements a no-operation evaluation loop.

    This class can be used to ignore the evaluation loop in an engine.
    """

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def eval(self, engine: BaseEngine) -> None:
        r"""Evaluates the model on the evaluation dataset.

        It is a no-operation method.

        Args:
            engine (``BaseEngine``): Specifies the engine.
        """



r"""This package contains the criterion base class and some implemented
criteria."""

__all__ = [
    "PackedSequenceLoss",
    "PaddedSequenceLoss",
    "VanillaLoss",
    "WeightedSumLoss",
]

from gravitorch.models.criterions.packed_seq import PackedSequenceLoss
from gravitorch.models.criterions.padded_seq import PaddedSequenceLoss
from gravitorch.models.criterions.vanilla import VanillaLoss
from gravitorch.models.criterions.weighted_sum import WeightedSumLoss
